Package: SimplicialComplex
Type: Package
Title: Topological Data Analysis: Simplicial Complex
Version: 0.1.0
Authors@R: c(person("ChiChien", "Wang", email = "kennywang2003@gmail.com", role = c("aut", "cre", "trl")))
Maintainer: ChiChien Wang <kennywang2003@gmail.com>
Description: Provides an implementation of simplicial complexes for 
    Topological Data Analysis (TDA). The package includes functions to 
    compute faces, boundary operators, Betti numbers, Euler characteristic, 
    and to construct simplicial complexes. It also implements persistent 
    homology, from building filtrations to computing persistence diagrams, 
    with the aim of helping readers understand the core concepts of 
    computational topology.
    Methods are based on standard references in persistent homology such as 
    Zomorodian and Carlsson (2005) <doi:10.1007/s00454-004-1146-y> and 
    Chazal and Michel (2021) <doi:10.3389/frai.2021.667963>.
Imports: Matrix, gtools, igraph, ggplot2
License: MIT + file LICENSE
URL: https://github.com/TDA-R/SimplicialComplex
BugReports: https://github.com/TDA-R/SimplicialComplex/issues
Encoding: UTF-8
RoxygenNote: 7.3.2
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2025-10-15 08:39:54 UTC; wangqiqian
Author: ChiChien Wang [aut, cre, trl]
Repository: CRAN
Date/Publication: 2025-10-20 19:30:15 UTC
Built: R 4.6.0; ; 2026-04-28 11:06:27 UTC; unix

Package: SimplicialComplex
Type: Package
Title: Topological Data Analysis: Simplicial Complex
Version: 0.1.2
Authors@R: c(person("ChiChien", "Wang", email = "kennywang2003@gmail.com", role = c("aut", "cre", "trl")))
Maintainer: ChiChien Wang <kennywang2003@gmail.com>
Description: Provides an implementation of simplicial complexes for
    Topological Data Analysis (TDA). The package includes functions to
    compute faces, boundary operators, Betti numbers, Euler characteristic,
    and to construct simplicial complexes, including Vietoris-Rips, Cech,
    Alpha, Delaunay, Witness, flood, and (via a Freudenthal triangulation)
    cubical complexes for grid and image data. It also implements persistent
    homology, from building filtrations (via a single build_filtration()
    entry point covering all of the above) to computing persistence
    diagrams, persistence landscapes, and Wasserstein/bottleneck distances
    between diagrams, with the aim of helping readers understand the core
    concepts of computational topology.
    Methods are based on standard references in persistent homology such as
    Zomorodian and Carlsson (2005) <doi:10.1007/s00454-004-1146-y>,
    Chazal and Michel (2021) <doi:10.3389/frai.2021.667963>, and
    Otter, Porter, Tillmann, Grindrod and Harrington (2017)
    <doi:10.1140/epjds/s13688-017-0109-5>.
Imports: Matrix, gtools, igraph, ggplot2, geometry, RANN, clue
License: MIT + file LICENSE
URL: https://github.com/TDA-R/SimplicialComplex
BugReports: https://github.com/TDA-R/SimplicialComplex/issues
Encoding: UTF-8
RoxygenNote: 7.3.2
Suggests: testthat (>= 3.0.0), torch
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-08-24 13:48:05 UTC; wangqiqian
Author: ChiChien Wang [aut, cre, trl]
Repository: CRAN
Date/Publication: 2026-08-24 18:40:50 UTC

# Generated by roxygen2: do not edit by hand

export(VietorisRipsComplex)
export(abstract_simplicial_complex)
export(betti_number)
export(boundary)
export(boundary_info)
export(build_vr_filtration)
export(euler_characteristic)
export(extract_persistence_pairs)
export(faces)
export(plot_persistence)
importFrom(Matrix,Matrix)
importFrom(Matrix,rankMatrix)
importFrom(Matrix,sparseMatrix)
importFrom(ggplot2,aes)
importFrom(ggplot2,coord_fixed)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_minimal)
importFrom(gtools,combinations)
importFrom(igraph,add_edges)
importFrom(igraph,graph.empty)
importFrom(igraph,max_cliques)
importFrom(stats,setNames)
importFrom(utils,combn)
importFrom(utils,tail)

# Topological Data Analysis: Simplicial Complex

[![CRAN status](https://www.r-pkg.org/badges/version/SimplicialComplex)](https://cran.r-project.org/package=SimplicialComplex) <a href="https://CRAN.R-project.org/package=SimplicialComplex" target="_blank" rel="noreferrer"> <img src="https://cranlogs.r-pkg.org/badges/grand-total/SimplicialComplex" alt="CRAN downloads" width="100" height="20"/> </a>

`SimplicialComplex` is a user-friendly Topological Data Analysis (TDA) package written entirely in R. While most TDA libraries (Dionysus, PHAT, GUDHI) are developed in Python and C++, implementing simplicial complexes natively in R makes them directly compatible with the rich ecosystem of statistical methods R already offers.

## Features

- **Simplicial complexes**: Build Vietoris–Rips, Alpha, Cech, Witness, Cubical, and Flood complexes from point clouds, or define abstract simplicial complexes by hand.
- **Topological invariants**: Faces, boundary matrices, Betti numbers, and the Euler characteristic.
- **Persistent homology**: filtrations, boundary-matrix reduction, persistence pairs, persistence diagrams, and persistence landscape.
- **Statistical**: Wasserstein distance & bottleneck distance were built in the latest version, with matching plot.
- **Examples**: Full worked examples in `inst/example`.

## Playground

Try the [interactive playground](https://tf3q5u-0-0.shinyapps.io/simplicialcomplex/) to get familiar with all the concepts used in TDA.

## References

- Zomorodian, A., & Carlsson, G. (2004). Computing persistent homology. *Proceedings of the Twentieth Annual Symposium on Computational Geometry*, 347–356.
- Chazal, F., & Michel, B. (2021). An introduction to topological data analysis: Fundamental and practical aspects for data scientists. *Frontiers in Artificial Intelligence*, 4, 667963.
- Graf, F., Pellizzoni, P., Uray, M., Huber, S., & Kwitt, R. (2025). The Flood Complex: Large-scale persistent homology on millions of points. *Advances in Neural Information Processing Systems*, 38.
- Otter, N., Porter, M. A., Tillmann, U., Grindrod, P., & Harrington, H. A. (2017). A roadmap for the computation of persistent homology. EPJ data science, 6(1), 17.

#' Compute the Euler characteristic \eqn{\chi} of a simplicial complex
#'
#' @param simplices A list of simplices (each a numeric vector).
#' @param tol Optional numerical tolerance to pass to \code{rankMatrix()}.
#' @return An integer representing the Euler characteristic \eqn{\chi}.
#'
#' @details
#' The Euler characteristic is computed as:
#' \deqn{\chi = \sum_{k=0}^{k_{\max}} (-1)^k \beta_k}
#' where \eqn{\beta_k} is the \eqn{k}th Betti number, and \eqn{k_{\max}} is the highest dimension of any simplex in the complex.
#'
#' Interpretation of values:
#' \itemize{
#'   \item \eqn{\chi = 2}: Sphere-like surfaces
#'   \item \eqn{\chi = 1}: Disk-like spaces
#'   \item \eqn{\chi = 0}: Torus-like or circle-like spaces
#'   \item \eqn{\chi < 0}: Surfaces with multiple handles or genus
#' }
#'
#' @seealso \code{\link{betti_number}}
#' @export
#' @examples
#' simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))
#' euler_characteristic(simplices, tol=0.1)
euler_characteristic <- function(
    simplices, tol
) {
  max_dim <- max(sapply(simplices, length))
  #　χ = Σ (-1)^k * β_k
  euler_char <- sum(sapply(0:(max_dim), function(bound_dim) {
    (-1)^bound_dim * betti_number(simplices, bound_dim, tol)
  }))

  return(euler_char)
}

library(Matrix)
library(gtools)
library(igraph)
library(ggplot2)
library(geometry)
library(clue)

source("./R/Faces.R")
source("./R/Betti.R")
source("./R/Boundary.R")
source("./R/EulerCharacteristic.R")
source("./R/ComplexUtils.R")
source("./R/AlphaComplex.R")
source("./R/CechComplex.R")
source("./R/FloodComplex.R") # generate_landmarks(), as_filtration()
source("./R/WitnessComplex.R")
source("./R/Filtration.R") # build_filtration()
source("./R/Persistence.R")
source("./R/PlotPD.R")

set.seed(42)
points <- matrix(runif(20), ncol = 2)

# Vietoris-Rips (existing)
vr_filtration <- build_filtration(points, method = "VR", eps_max = 1.2, max_dimension=2)
cat(sprintf("VR: %d simplices\n", length(vr_filtration)))

# Delaunay
delaunay_filtration <- build_filtration(points, method = "Delaunay", max_dimension=2)
cat(sprintf("Delaunay: %d simplices\n", length(delaunay_filtration)))

# Alpha
alpha_filtration <- build_filtration(points, method = "Alpha", eps_max = 1.2, max_dimension=2)
cat(sprintf("Alpha: %d simplices\n", length(alpha_filtration)))

# Cech
cech_filtration <- build_filtration(points, method = "Cech", eps_max = 1.2, max_dimension=2)
cech_max_dim <- max(lengths(lapply(cech_filtration, `[[`, "simplex")) - 1L)
cat(sprintf("Cech: %d simplices (max dimension %d)\n", length(cech_filtration), cech_max_dim))

# Witness
theta <- seq(0, 2 * pi, length.out = 40)[-40]
ring <- cbind(cos(theta), sin(theta)) + matrix(rnorm(2 * length(theta), sd = 0.03), ncol = 2)
n_landmarks <- 10
witness_filtration <- build_filtration(ring, method = "Witness", eps_max = 0.6, max_dimension=2, landmarks = n_landmarks)
cat(sprintf("Witness: %d simplices on %d landmarks\n", length(witness_filtration), n_landmarks))


show_top_bars <- function(pairs, label, top_n = 3) {
  pairs$pers <- pairs$death - pairs$birth
  cat(sprintf("\n--- %s ---\n", label))
  for (d in sort(unique(pairs$dim))) {
    pd <- pairs[pairs$dim == d, ]
    pd <- pd[order(-pd$pers), ]
    cat(sprintf("H%d (top %d of %d):\n", d, min(top_n, nrow(pd)), nrow(pd)))
    print(head(pd[, c("birth", "death", "pers")], top_n), row.names = FALSE)
  }
}

show_top_bars(persistence_pairs(vr_filtration), "VR")
show_top_bars(persistence_pairs(delaunay_filtration), "Delaunay")
show_top_bars(persistence_pairs(alpha_filtration), "Alpha")
show_top_bars(persistence_pairs(cech_filtration), "Cech")
show_top_bars(persistence_pairs(witness_filtration), "Witness")

plot_persistence(persistence_pairs(vr_filtration))

# Cubical
source("R/CubicalComplex.R")
image <- matrix(c(
  9, 9, 9, 9, 9,
  9, 1, 1, 1, 9,
  9, 1, 5, 1, 9,
  9, 1, 1, 1, 9,
  9, 9, 9, 9, 9
), nrow = 5, byrow = TRUE)

filtration <- build_cubical_filtration(image)
pairs <- persistence_pairs(filtration)
pairs[pairs$dim == 1, ]
plot_persistence(pairs)
show_top_bars(pairs, "Cubical")

library(Matrix)
library(gtools)
library(igraph)
library(ggplot2)

source("./R/Faces.R")
source("./R/Betti.R")
source("./R/Boundary.R")
source("./R/EulerCharacteristic.R")
source("R/ComplexUtils.R")
source("R/Filtration.R")

simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))

faces(simplices, target_dim=0)
faces(simplices, target_dim=1)
faces(simplices, target_dim=2)
faces(simplices, target_dim=3)

boundary(simplices, 0)
boundary(simplices, 1)
boundary(simplices, 2)

betti_number(simplices, 0, tol=0.1)
betti_number(simplices, 1, tol=0.1)
betti_number(simplices, 2, tol=0.1)
betti_number(simplices, 3, tol=0.1)

euler_characteristic(simplices, tol=0.1)

source("R/VRComplex.R")
set.seed(42)
points <- matrix(runif(20), ncol = 2)
epsilon <- 0.5

vr_complex <- VietorisRipsComplex(points, epsilon)
# cech_complex <- CechComplex(points, epsilon)

# E(vr_complex$network)
# vr_complex$simplices

faces(vr_complex$simplices, target_dim=2)
boundary(vr_complex$simplices, 2)
betti_number(vr_complex$simplices, 1, tol=0.1)

plot(
  vr_complex$network,
  layout = points,
  vertex.label = 1:nrow(points),
  vertex.size = 12,
  edge.arrow.mode = 0,
  asp = 1
)

filtration <- build_filtration(points, method = "VR", eps_max = 1, max_dimension=2)

source("R/Persistence.R")
res <- boundary_info(filtration)
pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)

source("R/PlotPD.R")
plot_persistence(pairs)

source("R/PersistenceLandscape.R")
landscape <- persistence_landscape(pairs, dimension = 0)
plot_landscape(landscape)

source("R/CubicalComplex.R")
image <- matrix(c(
  9, 9, 9, 9, 9,
  9, 1, 1, 1, 9,
  9, 1, 5, 1, 9,
  9, 1, 1, 1, 9,
  9, 9, 9, 9, 9
), nrow = 5, byrow = TRUE)

cubical_filtration <- build_cubical_filtration(image)
cubical_pairs <- persistence_pairs(cubical_filtration)
cubical_pairs[cubical_pairs$dim == 1, ] # H1: birth = 1 (ring closes), death = 5 (hole fills)
plot_persistence(cubical_pairs)

#' Restrict a filtration list to simplices up to a given dimension
#'
#' Drops every simplex of dimension greater than \code{max_dimension} from a
#' filtration list; everything else (order, ties, \code{t} values) is left
#' untouched.
#'
#' @param filist A filtration list, as produced by
#'   \code{\link{build_filtration}}/\code{\link{build_flood_filtration}}.
#' @param max_dimension Maximum simplex dimension to keep (\code{0} = vertices
#'   only, \code{1} = vertices + edges, etc.).
#' @return A filtration list containing only the entries with dimension
#'   \code{<= max_dimension}, in the same relative order.
#'
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
#' restrict_filtration(filtration, max_dimension = 1) # vertices + edges only
restrict_filtration <- function(filist, max_dimension) {
  Filter(function(x) length(x$simplex) - 1 <= max_dimension, filist)
}

#' Get the boundary matrix and its reduction information in matrix form
#'
#' @param filist Filtration list, each element includes simplex and time.
#' @param max_dimension Optional maximum homology dimension to eventually
#'   report via \code{\link{extract_persistence_pairs}}. When set,
#'   \code{filist} is first restricted to \code{max_dimension + 1} (one
#'   extra dimension, kept only so dimension-\code{max_dimension} classes
#'   are correctly killed - see \code{\link{restrict_filtration}}'s
#'   Details) before the boundary matrix is built and reduced. Leave
#'   \code{NULL} (default) to fall back to \code{filist}'s own
#'   \code{"max_dimension"} attribute (set automatically when it came from
#'   \code{\link{build_filtration}} with \code{max_dimension} supplied
#'   there), or to use \code{filist} uncapped if that attribute is also
#'   absent.
#' @return A list containing the boundary matrix, the last boundary row, the
#'   pivot owner for persistence extraction, and \code{filist} - the exact
#'   (possibly \code{max_dimension}-restricted) filtration list the other
#'   three elements were computed from, re-tagged with the same
#'   \code{"max_dimension"} attribute so \code{\link{extract_persistence_pairs}}
#'   can auto-detect it too. Always pass THIS \code{filist} back into
#'   \code{\link{extract_persistence_pairs}}, not your original one - see its
#'   Details for why.
#'
#' @importFrom utils combn tail
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
#' res <- boundary_info(filtration)
boundary_info <- function(filist, max_dimension = NULL) {
  if (is.null(max_dimension)) max_dimension <- attr(filist, "max_dimension")
  if (!is.null(max_dimension)) {
    filist <- restrict_filtration(filist, max_dimension + 1)
    # Filter()/subsetting on a plain list drops attributes, so re-tag here -
    # this is what lets extract_persistence_pairs(res$filist, ...) auto-
    # detect max_dimension without it being passed again.
    attr(filist, "max_dimension") <- max_dimension
  }

  boundary <- matrix(nrow = length(filist), ncol = length(filist), data = 0)
  name_vec = list()
  metadata = list()

  for (i in seq_along(filist)) {
    # add filist[[i]]$simplex to nam_vec
    name_vec[[i]] <- filist[[i]]$simplex
    # add info
    info_matrix <- combn(filist[[i]]$simplex, length(filist[[i]]$simplex) -1)
    metadata[[i]] <- lapply(seq_len(ncol(info_matrix)), function(j) info_matrix[, j])
  }

  rownames(boundary) <- colnames(boundary) <- sapply(name_vec, function(x) paste(x, collapse = " "))

  # CHANGED (performance): O(1) name -> row-index lookup instead of scanning
  # rownames with which(rownames(boundary) == row_name) for every face,
  # which made the fill-in O(n^2) string comparisons.
  row_index <- stats::setNames(seq_along(filist), rownames(boundary))

  last_1 <- rep(NA, ncol(boundary))
  for(i in seq_along(filist)) {

    # create boundary matrix
    dt <- metadata[[i]]
    for (one_face in dt) {
      if (length(one_face) == 0) next  # vertices have an empty boundary
      row_idx <- row_index[paste(one_face, collapse = " ")]
      if (is.na(row_idx)) next         # face not in the filtration (as before)
      boundary[row_idx, i] <- 1
    }

    # last boundary for reduction
    for_row <- boundary[, i]
    all_1 <- which(for_row == 1)
    if(length(all_1) > 0) {
      last_1[i] <- tail(all_1, 1)
    }
  }

  xor_vec <- function(a, b) {
    (a + b) %% 2
  }

  # record which row is used by which column
  pivot_owner <- rep(NA, nrow(boundary))
  for(j in seq_along(filist)) {
    repeat {

      low_idx <- last_1[j] # last 1 row idx
      if(is.na(low_idx)) break
      owner_col <- pivot_owner[low_idx]

      if(is.na(owner_col)) {
        pivot_owner[low_idx] <- j
        break
      } else {
        # XOR owner_col to j
        boundary[, j] <- xor_vec(boundary[, j], boundary[, owner_col])
        # update last_1[j]
        for_row <- boundary[, j]
        all_1 <- which(for_row == 1)

        if(length(all_1) > 0) {
          last_1[j] <- tail(all_1, 1)
        } else {
          last_1[j] <- NA
        }
      }
    }
  }

  return(list(boundary = boundary, last_1 = last_1, pivot_owner = pivot_owner, filist = filist))

}

#' This function extracts the persistence from combining the boundary matrix and its filtration
#'
#' @param filist Filtration list, each element includes simplex and time.
#'   When \code{boundary_info()} was called with \code{max_dimension} set,
#'   this MUST be \code{res$filist} (the restricted list it returned), not
#'   your original filtration - \code{last_1}/\code{pivot_owner} are indexed
#'   positionally against whatever \code{filist} \code{boundary_info()}
#'   actually used, so passing a different-length \code{filist} here would
#'   silently pair the wrong simplices. See Details.
#' @param last_1 The last 1 row index for each column in boundary matrix (after reduction).
#' @param pivot_owner The column index owning the pivot row.
#' @param max_dimension Optional maximum homology dimension to report. Set
#'   this to the SAME value passed to \code{boundary_info()} (which kept one
#'   extra dimension internally for correct killers - see
#'   \code{\link{restrict_filtration}}'s Details); only that extra dimension
#'   is dropped here. Leave \code{NULL} (default) to fall back to
#'   \code{filist}'s own \code{"max_dimension"} attribute (present when
#'   \code{filist} is \code{res$filist} from a \code{boundary_info()} call
#'   that used \code{max_dimension}, directly or via its own fallback to
#'   \code{\link{build_filtration}}'s attribute), or to report every
#'   dimension present in \code{filist} if that attribute is also absent.
#' @return A data frame with columns: dimension, birth, and death.
#'
#' @details
#' \code{boundary_info()} and \code{extract_persistence_pairs()} are two
#' halves of one computation - \code{last_1}/\code{pivot_owner} only mean
#' anything relative to the exact \code{filist} \code{boundary_info()} used
#' internally. Whenever \code{max_dimension} is involved, always call as:
#' \preformatted{
#' res <- boundary_info(filtration, max_dimension = k)
#' pairs <- extract_persistence_pairs(res$filist, res$last_1, res$pivot_owner,
#'                                     max_dimension = k)
#' }
#' A length mismatch between \code{filist} and \code{last_1}/\code{pivot_owner}
#' is refused with an error rather than silently producing wrong pairs - see
#' \code{\link{persistence_pairs}} for a one-call alternative that cannot run
#' into this.
#'
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
#' res <- boundary_info(filtration)
#' pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
extract_persistence_pairs <- function(filist, last_1, pivot_owner, max_dimension = NULL) {
  if (is.null(max_dimension)) max_dimension <- attr(filist, "max_dimension")
  if (length(filist) != length(last_1) || length(filist) != length(pivot_owner)) {
    stop(sprintf(
      paste(
        "filist has %d simplices but last_1/pivot_owner have %d/%d - these",
        "must come from the same boundary_info() call. If boundary_info()",
        "was called with max_dimension set, pass its res$filist here, not",
        "your original filtration."
      ),
      length(filist), length(last_1), length(pivot_owner)
    ))
  }

  pairs <- list()
  for (row in seq_along(pivot_owner)) {
    col <- pivot_owner[row]
    if (!is.na(col)) {
      # row: the simplex being killed, col: the simplex killing it
      birth_simplex <- filist[[row]]
      death_simplex <- filist[[col]]

      birth_time <- birth_simplex$t
      death_time <- death_simplex$t
      # the dimension of simplex that is killed
      dim <- length(birth_simplex$simplex) - 1

      pairs[[length(pairs)+1]] <- list(
        dim = dim,
        birth = birth_time,
        death = death_time,
        birth_simplex = birth_simplex$simplex,
        death_simplex = death_simplex$simplex
      )
    }
  }

  # CHANGED (bug fix): essential classes.
  #
  # Old condition:  !(i %in% pivot_owner)
  # The *values* of pivot_owner are killer COLUMN indices, so this only
  # excluded negative (killing) simplices. A positive simplex that was
  # already paired (its ROW is a pivot, i.e. !is.na(pivot_owner[i]))
  # slipped through and produced a spurious infinite bar. Example: for the
  # 4-point square, vertices 2..4 die at t = 1 but were also reported as
  # essential H0 classes.
  #
  # Correct definition: a simplex i is essential iff
  #   (a) its column reduces to zero  -> is.na(last_1[i])   (positive simplex)
  #   (b) its row is never a pivot    -> is.na(pivot_owner[i]) (never killed)
  for (i in seq_along(filist)) {
    if (is.na(last_1[i]) && is.na(pivot_owner[i]) &&
        !all(is.na(filist[[i]]$simplex))) {
      birth_time <- filist[[i]]$t
      dim <- length(filist[[i]]$simplex) - 1
      pairs[[length(pairs)+1]] <- list(
        dim = dim,
        birth = birth_time,
        death = Inf,
        birth_simplex = filist[[i]]$simplex,
        death_simplex = NA
      )
    }
  }

  df <- data.frame(
    dim = sapply(pairs, function(x) x$dim),
    birth = sapply(pairs, function(x) x$birth),
    death = sapply(pairs, function(x) x$death)
  )

  if (!is.null(max_dimension)) {
    # drop the extra dimension boundary_info() kept only so max_dimension's
    # classes were correctly killed above - never meant to be reported.
    df <- df[df$dim <= max_dimension, , drop = FALSE]
    rownames(df) <- NULL
  }

  return(df)
}

#' Persistence pairs via sparse boundary reduction (for large filtrations)
#'
#' Same standard column-reduction algorithm as
#' \code{boundary_info()} + \code{extract_persistence_pairs()}, but each
#' column is stored as a sorted integer vector over GF(2) instead of a row of
#' a dense n x n matrix. Memory drops from O(n^2) to O(total number of
#' non-zeros), which is what makes filtrations with thousands to hundreds of
#' thousands of simplices (e.g. Flood or large Vietoris-Rips complexes)
#' feasible. Output is identical to the dense pipeline.
#'
#' @param filist Filtration list, each element includes simplex and time.
#' @param max_dimension Optional maximum homology dimension to report
#'   (\code{0} = H0 only, \code{1} = H0 and H1, etc.). When set, one extra
#'   dimension is kept internally so dimension-\code{max_dimension} classes
#'   still get their correct death time from their true (max_dimension+1)
#'   killers, and only that extra dimension is dropped from the returned
#'   pairs - see \code{\link{restrict_filtration}}'s Details for why a plain
#'   truncation would be wrong. Leave \code{NULL} (default) to fall back to
#'   \code{filist}'s own \code{"max_dimension"} attribute (set automatically
#'   when \code{filist} came from \code{\link{build_filtration}} with
#'   \code{max_dimension} supplied there), or to report every dimension
#'   present in \code{filist} if that attribute is also absent.
#' @return A data frame with columns: dim, birth, and death.
#'
#' @importFrom utils combn
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
#' pairs <- persistence_pairs(filtration)
#' pairs_h0_only <- persistence_pairs(filtration, max_dimension = 0)
#' # or bake the cap into the filtration itself, and drop the argument here:
#' capped <- build_filtration(points, method = "VR", eps_max = 1.2,
#'                             max_dimension = 0)
#' pairs_h0_only2 <- persistence_pairs(capped)
persistence_pairs <- function(filist, max_dimension = NULL) {
  if (is.null(max_dimension)) max_dimension <- attr(filist, "max_dimension")
  if (!is.null(max_dimension)) {
    filist <- restrict_filtration(filist, max_dimension + 1)
  }
  n <- length(filist)
  keys <- vapply(filist, function(x) paste(x$simplex, collapse = " "), "")
  index <- new.env(hash = TRUE, parent = emptyenv())
  for (i in seq_len(n)) assign(keys[i], i, envir = index)

  # sparse boundary columns: indices of the facets of each simplex
  cols <- vector("list", n)
  for (i in seq_len(n)) {
    s <- filist[[i]]$simplex
    if (length(s) == 1L) { cols[[i]] <- integer(0); next }
    fmat <- combn(s, length(s) - 1L)
    cols[[i]] <- sort(vapply(seq_len(ncol(fmat)), function(j)
      get(paste(fmat[, j], collapse = " "), envir = index), 0L))
  }

  # XOR of two sparse GF(2) columns = symmetric difference of index sets
  symdiff <- function(a, b) sort.int(c(a[!(a %in% b)], b[!(b %in% a)]))

  pivot_owner <- rep(NA_integer_, n)
  for (j in seq_len(n)) {
    col <- cols[[j]]
    repeat {
      if (length(col) == 0L) break
      piv <- col[length(col)]           # lowest 1 = last index
      owner <- pivot_owner[piv]
      if (is.na(owner)) { pivot_owner[piv] <- j; break }
      col <- symdiff(col, cols[[owner]])
    }
    cols[[j]] <- col
  }

  dims <- lengths(lapply(filist, `[[`, "simplex")) - 1L
  ts <- vapply(filist, `[[`, 0, "t")

  killed <- which(!is.na(pivot_owner))
  pairs <- data.frame(
    dim = dims[killed],
    birth = ts[killed],
    death = ts[pivot_owner[killed]]
  )
  # essential classes: zero column (positive simplex) that is never paired
  zero_cols <- which(vapply(cols, length, 0L) == 0L)
  essential <- zero_cols[is.na(pivot_owner[zero_cols])]
  if (length(essential)) {
    pairs <- rbind(pairs, data.frame(
      dim = dims[essential], birth = ts[essential], death = Inf))
  }
  rownames(pairs) <- NULL
  pairs <- pairs[order(pairs$dim, pairs$birth), ]

  if (!is.null(max_dimension)) {
    # drop the extra dimension kept only so max_dimension's classes were
    # correctly killed above - it was never meant to be reported itself.
    pairs <- pairs[pairs$dim <= max_dimension, , drop = FALSE]
    rownames(pairs) <- NULL
  }
  pairs
}

library(testthat)
library(SimplicialComplex)

test_that("faces works correctly", {
  simplices <- list(c(2, 1, 3), c(4, 2), c(5), c(2, 3, 5, 4))
  expect_length(faces(simplices, target_dim=0), 5)
})

test_that("persistence_landscape computes the tent-function landscape correctly", {
  # two identical bars (0, 2): tent peaks at t = 1 with height 1
  df <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(2, 2))
  landscape <- persistence_landscape(df, dimension = 0, resolution = 3, t_range = c(0, 2))

  at_0 <- landscape[landscape$t == 0, ]
  at_1 <- landscape[landscape$t == 1, ]
  at_2 <- landscape[landscape$t == 2, ]

  expect_setequal(landscape$k, c(1, 2))
  expect_equal(at_0$value, c(0, 0))
  expect_equal(at_1$value, c(1, 1))
  expect_equal(at_2$value, c(0, 0))
})

test_that("persistence_landscape drops essential (death = Inf) pairs", {
  df <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, Inf))
  expect_message(
    landscape <- persistence_landscape(df, dimension = 0, resolution = 5),
    "essential"
  )
  expect_true(all(is.finite(landscape$value)))
  expect_equal(max(landscape$k), 1)
})

test_that("build_filtration(method = 'VR') matches VietorisRipsComplex directly", {
  points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
  filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
  vr <- VietorisRipsComplex(points, epsilon = 1.2)
  expect_equal(length(filtration), sum(sapply(0:(max(sapply(vr$simplices, length)) - 1),
                                               function(k) length(faces(vr$simplices, k)))))
  expect_true(all(vapply(filtration, function(x) x$t >= 0, TRUE)))
  expect_true(is.unsorted(sapply(filtration, `[[`, "t")) == FALSE)
})

test_that("build_filtration requires eps_max for VR/Cech/Witness but not Delaunay", {
  points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
  expect_error(build_filtration(points, method = "VR"), "eps_max")
  expect_error(build_filtration(points, method = "Cech"), "eps_max")
  expect_error(build_filtration(points, method = "Witness"), "eps_max")
  expect_error(build_filtration(points, method = "Witness", eps_max = 1), "landmarks")
  expect_silent(build_filtration(points, method = "Delaunay"))
})

test_that("AlphaComplex reproduces the textbook unit-square example", {
  # 4 corners of the unit square: the hole (empty square) is born once the
  # perimeter closes (edge length 1 -> alpha value 0.5) and dies once the
  # diagonal enters (length sqrt(2) -> alpha value sqrt(2)/2)
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1), ncol = 2, byrow = TRUE)
  filtration <- build_filtration(points, method = "Alpha")
  pairs <- persistence_pairs(filtration)
  h1 <- pairs[pairs$dim == 1, ]
  # simultaneous events at t = sqrt(2)/2 (the diagonal edge and both
  # triangles all enter at once) can also produce a zero-persistence bar;
  # only the genuine (birth < death) feature matters here
  real <- h1[h1$death > h1$birth, ]
  expect_equal(nrow(real), 1)
  expect_equal(real$birth, 0.5, tolerance = 1e-8)
  expect_equal(real$death, sqrt(2) / 2, tolerance = 1e-8)
})

test_that("DelaunayComplex is AlphaComplex at epsilon = Inf", {
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1, 0.5, 0.4), ncol = 2, byrow = TRUE)
  dc <- DelaunayComplex(points)
  ac <- AlphaComplex(points, epsilon = Inf)
  expect_equal(dc$simplices, ac$simplices)
  expect_equal(dc$filtration, ac$filtration)
  expect_true(inherits(dc, "delaunay_complex"))
})

test_that("CechComplex includes the full simplex once epsilon reaches the circumradius", {
  # 4 corners of the unit square: every point is within sqrt(2)/2 of the
  # center, so the whole 4-point set enters the Cech complex at exactly that
  # scale - the well-known "Cech complex can jump straight to high
  # dimension" behaviour that distinguishes it from Alpha/Delaunay.
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1), ncol = 2, byrow = TRUE)
  filtration <- build_filtration(points, method = "Cech", eps_max = 2)
  full <- Filter(function(x) length(x$simplex) == 4, filtration)
  expect_length(full, 1)
  expect_equal(full[[1]]$t, sqrt(2) / 2, tolerance = 1e-8)
})

test_that("restrict_filtration keeps only simplices at or below the given dimension", {
  simplices <- list(c(1), c(2), c(1, 2), c(1, 2, 3))
  filist <- list(
    list(simplex = c(1), t = 0), list(simplex = c(2), t = 0),
    list(simplex = c(1, 2), t = 0.5), list(simplex = c(1, 2, 3), t = 1)
  )
  r0 <- restrict_filtration(filist, max_dimension = 0)
  expect_true(all(vapply(r0, function(x) length(x$simplex) - 1, 0) <= 0))
  r1 <- restrict_filtration(filist, max_dimension = 1)
  expect_true(all(vapply(r1, function(x) length(x$simplex) - 1, 0) <= 1))
  expect_length(r1, 3)
})

test_that("persistence_pairs(max_dimension=) reports correct death times, not a naive drop", {
  # the Cech blow-up example above: H1's one real feature (birth 0.5) is
  # killed by a triangle (a dimension-2 simplex) entering at sqrt(2)/2 - so
  # capping at max_dimension = 1 must still report death = sqrt(2)/2 exactly
  # as the uncapped computation does, by keeping dimension 2 internally as a
  # potential killer and only dropping it from the *output*. A naive
  # dim <= 1 truncation (no "+1") would instead leave that class essential.
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1), ncol = 2, byrow = TRUE)
  filtration <- build_filtration(points, method = "Cech", eps_max = 1.2)

  full <- persistence_pairs(filtration)
  capped <- persistence_pairs(filtration, max_dimension = 1)
  naive <- persistence_pairs(restrict_filtration(filtration, 1))

  expect_true(all(capped$dim <= 1))

  h1_full <- full[full$dim == 1 & full$death > full$birth, c("birth", "death")]
  h1_capped <- capped[capped$dim == 1 & capped$death > capped$birth, c("birth", "death")]
  rownames(h1_full) <- NULL; rownames(h1_capped) <- NULL
  expect_equal(h1_capped, h1_full) # correct death time preserved, not dropped

  h1_naive <- naive[naive$dim == 1 & naive$birth == 0.5, ]
  expect_true(is.infinite(h1_naive$death)) # demonstrates the naive approach is wrong
})

test_that("flood_persistence(max_dimension=) applies the same correction", {
  set.seed(5)
  pts <- matrix(rnorm(60), ncol = 2)
  filtration <- build_flood_filtration(pts, landmarks = 12, points_per_edge = 6)
  full <- flood_persistence(filtration)
  capped <- flood_persistence(filtration, max_dimension = 0)
  expect_true(all(capped$dim == 0))
  h0_full <- full[full$dim == 0, c("birth", "death")]
  h0_capped <- capped[capped$dim == 0, c("birth", "death")]
  rownames(h0_full) <- NULL; rownames(h0_capped) <- NULL
  expect_equal(h0_capped, h0_full)
})

test_that("boundary_info()/extract_persistence_pairs() honour max_dimension identically to persistence_pairs()", {
  # same Cech blow-up example as the persistence_pairs() max_dimension test:
  # the dense pipeline (boundary_info + extract_persistence_pairs) must
  # apply the exact same "+1 internally, trim the output" correction as the
  # sparse persistence_pairs(), so the two agree bit for bit.
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1), ncol = 2, byrow = TRUE)
  filtration <- build_filtration(points, method = "Cech", eps_max = 1.2)

  res <- boundary_info(filtration, max_dimension = 1)
  dense <- extract_persistence_pairs(res$filist, res$last_1, res$pivot_owner, max_dimension = 1)
  sparse <- persistence_pairs(filtration, max_dimension = 1)

  expect_true(all(dense$dim <= 1))
  d1 <- dense[order(dense$dim, dense$birth), c("dim", "birth", "death")]
  d2 <- sparse[order(sparse$dim, sparse$birth), c("dim", "birth", "death")]
  rownames(d1) <- NULL; rownames(d2) <- NULL
  expect_equal(d1, d2)
})

test_that("extract_persistence_pairs() refuses a filist that doesn't match last_1/pivot_owner", {
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1), ncol = 2, byrow = TRUE)
  filtration <- build_filtration(points, method = "Cech", eps_max = 1.2)
  res <- boundary_info(filtration, max_dimension = 1) # res$filist is shorter than filtration

  expect_error(
    extract_persistence_pairs(filtration, res$last_1, res$pivot_owner, max_dimension = 1),
    "must come from the same boundary_info"
  )
  # using the matching filist works fine
  expect_silent(extract_persistence_pairs(res$filist, res$last_1, res$pivot_owner, max_dimension = 1))
})

test_that("boundary_info()/extract_persistence_pairs() without max_dimension are unchanged", {
  points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
  filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
  res <- boundary_info(filtration)
  expect_identical(res$filist, filtration)
  pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
  sparse <- persistence_pairs(filtration)
  p1 <- pairs[order(pairs$dim, pairs$birth), c("dim", "birth", "death")]
  p2 <- sparse[order(sparse$dim, sparse$birth), c("dim", "birth", "death")]
  rownames(p1) <- NULL; rownames(p2) <- NULL
  expect_equal(p1, p2)
})

test_that("build_filtration(max_dimension=) tags its result so downstream functions need no parameter", {
  # Same Cech blow-up example as the persistence_pairs(max_dimension=) test
  # above, but the cap now comes from build_filtration() itself - the
  # downstream calls below pass no max_dimension at all and must still match
  # the already-verified explicit-parameter path exactly.
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1), ncol = 2, byrow = TRUE)
  full <- build_filtration(points, method = "Cech", eps_max = 1.2)
  capped <- build_filtration(points, method = "Cech", eps_max = 1.2, max_dimension = 1)

  expect_equal(attr(capped, "max_dimension"), 1)
  # the "+1 trick": dimension 2 is still present internally...
  expect_true(any(vapply(capped, function(x) length(x$simplex) - 1, 0) == 2))
  # ...but nothing above it is built at all, unlike the uncapped filtration
  expect_true(any(vapply(full, function(x) length(x$simplex) - 1, 0) > 2) ||
                max(vapply(full, function(x) length(x$simplex) - 1, 0)) == 2)
  expect_true(length(capped) <= length(full))

  explicit <- persistence_pairs(full, max_dimension = 1)
  auto <- persistence_pairs(capped) # no max_dimension argument at all
  e <- explicit[order(explicit$dim, explicit$birth), c("dim", "birth", "death")]
  a <- auto[order(auto$dim, auto$birth), c("dim", "birth", "death")]
  rownames(e) <- NULL; rownames(a) <- NULL
  expect_equal(a, e)
})

test_that("build_filtration(max_dimension=) auto-detection also reaches boundary_info()/extract_persistence_pairs()", {
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1), ncol = 2, byrow = TRUE)
  capped <- build_filtration(points, method = "Cech", eps_max = 1.2, max_dimension = 1)

  res <- boundary_info(capped) # no max_dimension argument
  expect_equal(attr(res$filist, "max_dimension"), 1)
  dense <- extract_persistence_pairs(res$filist, res$last_1, res$pivot_owner) # none here either
  sparse <- persistence_pairs(capped)

  expect_true(all(dense$dim <= 1))
  d1 <- dense[order(dense$dim, dense$birth), c("dim", "birth", "death")]
  d2 <- sparse[order(sparse$dim, sparse$birth), c("dim", "birth", "death")]
  rownames(d1) <- NULL; rownames(d2) <- NULL
  expect_equal(d1, d2)
})

test_that("an explicit downstream max_dimension overrides build_filtration()'s own cap", {
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1), ncol = 2, byrow = TRUE)
  capped <- build_filtration(points, method = "Cech", eps_max = 1.2, max_dimension = 1)

  narrower <- persistence_pairs(capped, max_dimension = 0)
  expect_true(all(narrower$dim == 0))

  res <- boundary_info(capped, max_dimension = 0)
  expect_equal(attr(res$filist, "max_dimension"), 0)
  dense <- extract_persistence_pairs(res$filist, res$last_1, res$pivot_owner, max_dimension = 0)
  expect_true(all(dense$dim == 0))
})

test_that("build_filtration() without max_dimension carries no attribute and is unaffected", {
  points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
  filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
  expect_null(attr(filtration, "max_dimension"))

  res <- boundary_info(filtration)
  expect_null(attr(res$filist, "max_dimension"))
  expect_identical(res$filist, filtration)
})

test_that("build_filtration(max_dimension=) restricts Alpha/Delaunay complexes correctly (post-hoc branch)", {
  # Alpha/Delaunay have no build-time dimension cap, so build_filtration()
  # restricts them after the fact - a different code path from the VR/Cech/
  # Witness branches (which cap simplices_to_filtration()'s kmax directly).
  # Exercise it explicitly so both branches are covered.
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1, 0.5, 0.5), ncol = 2, byrow = TRUE)

  for (method in c("Alpha", "Delaunay")) {
    full <- build_filtration(points, method = method)
    capped <- build_filtration(points, method = method, max_dimension = 1)
    expect_equal(attr(capped, "max_dimension"), 1)
    expect_true(max(vapply(capped, function(x) length(x$simplex) - 1, 0)) <= 2)

    explicit <- persistence_pairs(full, max_dimension = 1)
    auto <- persistence_pairs(capped)
    e <- explicit[order(explicit$dim, explicit$birth), c("dim", "birth", "death")]
    a <- auto[order(auto$dim, auto$birth), c("dim", "birth", "death")]
    rownames(e) <- NULL; rownames(a) <- NULL
    expect_equal(a, e)
  }
})

test_that("WitnessComplex builds a valid complex on the requested number of landmarks", {
  set.seed(7)
  theta <- seq(0, 2 * pi, length.out = 25)[-25]
  points <- cbind(cos(theta), sin(theta))
  filtration <- build_filtration(points, method = "Witness", eps_max = 0.6, landmarks = 8)
  vertex_ids <- unique(unlist(lapply(filtration, function(x) x$simplex[length(x$simplex) == 1])))
  expect_true(length(unique(unlist(lapply(filtration[sapply(filtration, function(x) length(x$simplex) == 1)],
                                           `[[`, "simplex")))) <= 8)
  expect_true(all(vapply(filtration, function(x) x$t >= 0, TRUE)))
})

test_that("wasserstein_distance and bottleneck_distance are zero for identical diagrams", {
  df <- data.frame(dim = c(0, 0, 1), birth = c(0, 0, 0.2), death = c(1, 2, 0.9))
  expect_equal(wasserstein_distance(df, df, dimension = 0, p = 2), 0)
  expect_equal(bottleneck_distance(df, df, dimension = 0), 0)
  expect_equal(wasserstein_distance(df, df, dimension = 1, p = 1), 0)
})

test_that("wasserstein_distance and bottleneck_distance match hand-computed values", {
  df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
  df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
  # optimal matching pairs (0,1)-(0,1) exactly and (0,2)-(0,3) at distance 1
  expect_equal(wasserstein_distance(df1, df2, dimension = 0, p = 2), 1, tolerance = 1e-8)
  expect_equal(bottleneck_distance(df1, df2, dimension = 0), 1, tolerance = 1e-8)
})

test_that("distance functions cap essential pairs by birth time instead of dropping them", {
  # df1 has an essential class (0, Inf); dropping it would let it vanish
  # from the comparison entirely. It should instead be capped (at 2x the
  # largest finite value present, i.e. death = 2 here) and matched by its
  # birth time like any other point.
  df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, Inf))
  df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 1))
  expect_message(w <- wasserstein_distance(df1, df2, dimension = 0, p = 2), "Capping")
  expect_equal(w, 1, tolerance = 1e-8)
  expect_message(b <- bottleneck_distance(df1, df2, dimension = 0), "Capping")
  expect_equal(b, 1, tolerance = 1e-8)
})

test_that("essential pairs still give zero self-distance", {
  df <- data.frame(dim = c(0, 1), birth = c(0, 0.2), death = c(Inf, Inf))
  expect_equal(wasserstein_distance(df, df, dimension = 0, p = 2), 0)
  expect_equal(bottleneck_distance(df, df, dimension = 1), 0)
})

test_that("diagram_matching's distance agrees with wasserstein_distance/bottleneck_distance", {
  df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
  df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))

  mw <- diagram_matching(df1, df2, dimension = 0, distance = "wasserstein", p = 2)
  expect_equal(mw$distance, wasserstein_distance(df1, df2, dimension = 0, p = 2), tolerance = 1e-8)
  # both points matched real-to-real: (0,1)-(0,1) at zero cost, (0,2)-(0,3) at cost 1
  expect_equal(nrow(mw$matches), 2)
  expect_true(all(mw$matches$type == "real-real"))

  mb <- diagram_matching(df1, df2, dimension = 0, distance = "bottleneck")
  expect_equal(mb$distance, bottleneck_distance(df1, df2, dimension = 0), tolerance = 1e-8)
})

test_that("diagram_matching caps essential pairs and flags them in the matching table", {
  df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, Inf))
  df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 1))

  expect_message(m <- diagram_matching(df1, df2, dimension = 0, distance = "wasserstein", p = 2),
                  "Capping")
  expect_equal(m$distance, 1, tolerance = 1e-8)
  expect_true(any(m$matches$x_essential, na.rm = TRUE))
  # the essential point's capped death matches cap_essential's own convention (2x max finite)
  essential_row <- m$matches[which(m$matches$x_essential), ]
  expect_equal(essential_row$x_death, 2, tolerance = 1e-8)
})

test_that("diagram_matching errors when neither diagram has points in the requested dimension", {
  df1 <- data.frame(dim = c(0), birth = c(0), death = c(1))
  df2 <- data.frame(dim = c(0), birth = c(0), death = c(1))
  expect_error(diagram_matching(df1, df2, dimension = 1), "nothing to match")
})

test_that("plot_matching returns a ggplot object built from the same matching", {
  df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
  df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
  p <- plot_matching(df1, df2, dimension = 0, distance = "wasserstein", p = 2,
                      labels = c("A", "B"))
  expect_s3_class(p, "ggplot")
  expect_true(grepl("distance = 1", p$labels$title))
})

test_that("compare_complexes summarizes every requested method", {
  points <- matrix(c(0, 0, 1, 0, 0, 1, 1, 1, 0.5, 0.5), ncol = 2, byrow = TRUE)
  cmp <- compare_complexes(points, methods = c("VR", "Delaunay", "Alpha"), eps_max = 1.2)
  expect_setequal(cmp$summary$method, c("VR", "Delaunay", "Alpha"))
  expect_true(all(cmp$summary$n_simplices > 0))
  expect_named(cmp$diagrams, c("VR", "Delaunay", "Alpha"))
})

test_that("complex_distance computes a symmetric-sensible distance and validates inputs", {
  set.seed(3)
  cloud_a <- matrix(rnorm(24), ncol = 2)
  cloud_b <- cloud_a + 0.05
  res <- complex_distance(cloud_a, cloud_b, method = "VR", distance = "bottleneck",
                           dimension = 0, eps_max = 0.6)
  expect_true(res$distance >= 0)
  self <- complex_distance(cloud_a, cloud_a, method = "Delaunay", distance = "wasserstein",
                            dimension = 1, p = 2)
  expect_equal(self$distance, 0)
  expect_error(complex_distance(cloud_a, cloud_b, method = "Cech"), "eps_max")
})

test_that("build_cubical_filtration finds the hole in a ring image", {
  # ring (1) around a hole (5) around a background (9): the hole should be
  # born when the ring closes (t = 1) and die when its center fills (t = 5)
  image <- matrix(c(
    9, 9, 9, 9, 9,
    9, 1, 1, 1, 9,
    9, 1, 5, 1, 9,
    9, 1, 1, 1, 9,
    9, 9, 9, 9, 9
  ), nrow = 5, byrow = TRUE)

  filtration <- build_cubical_filtration(image)
  pairs <- persistence_pairs(filtration)
  h1 <- pairs[pairs$dim == 1, ]

  expect_true(any(h1$birth == 1 & h1$death == 5))
  expect_true(all(vapply(filtration, function(x) length(x$simplex), 0) <= 3))
})

test_that("plot_persistence labels dimensions discretely as H0/H1/H2/... with a legend", {
  df <- data.frame(dim = c(0, 0, 1, 2), birth = c(0, 0, 0.5, 0.7), death = c(0.5, Inf, 0.7, 0.9))
  p <- plot_persistence(df)

  expect_s3_class(p, "ggplot")
  expect_true(is.factor(p$data$dim_label))
  expect_equal(levels(p$data$dim_label), c("H0", "H1", "H2"))
  expect_false(identical(p$theme$legend.position, "none"))
})

#' Compute the persistence landscape of a persistence diagram
#'
#' Converts a persistence diagram data frame (the direct output of
#' \code{extract_persistence_pairs()}, \code{persistence_pairs()}, or
#' \code{flood_persistence()}) into its persistence landscape: a collection of
#' continuous, piecewise-linear functions \eqn{\lambda(k, \cdot)}, obtained by
#' overlaying the "tent" function of every birth-death pair and, at each time
#' \code{t}, taking the \eqn{k}th largest tent value (Bubenik, 2015; Chazal
#' and Michel, 2021, Section 5.4).
#'
#' @param df A persistence diagram data frame with columns \code{dim},
#'   \code{birth}, \code{death} (e.g. the output of
#'   \code{extract_persistence_pairs()}).
#' @param dimension Homological dimension to extract the landscape for.
#' @param k_max Number of landscape levels \eqn{\lambda(1,\cdot), \dots,
#'   \lambda(k_{max},\cdot)} to return. Defaults to all levels supported by
#'   the diagram (the number of finite birth-death pairs in \code{dimension});
#'   levels beyond that are identically zero.
#' @param resolution Number of grid points used to discretize each landscape
#'   function.
#' @param t_range Optional \code{c(t_min, t_max)} grid range. Defaults to
#'   \code{c(min(birth), max(death))} over the selected pairs.
#' @return A data frame with columns \code{t}, \code{k}, \code{value}: the
#'   discretized landscape functions, one row per (grid point, level) pair.
#'
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
#' res <- boundary_info(filtration)
#' pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
#' landscape <- persistence_landscape(pairs, dimension = 0)
persistence_landscape <- function(
    df, dimension = 0, k_max = NULL, resolution = 500, t_range = NULL
) {

  sub <- df[df$dim == dimension, , drop = FALSE]

  n_inf <- sum(is.infinite(sub$death))
  if (n_inf > 0) {
    message(sprintf(
      "Dropping %d essential (death = Inf) pair(s) in dimension %d; persistence landscapes are only defined for finite bars.",
      n_inf, dimension
    ))
    sub <- sub[is.finite(sub$death), , drop = FALSE]
  }

  if (nrow(sub) == 0) {
    stop(sprintf("No finite birth-death pairs found in dimension %d.", dimension))
  }

  b <- sub$birth
  d <- sub$death
  n <- length(b)

  if (is.null(t_range)) {
    t_min <- min(b)
    t_max <- max(d)
  } else {
    t_min <- t_range[1]
    t_max <- t_range[2]
  }
  t_seq <- seq(t_min, t_max, length.out = resolution)

  if (is.null(k_max)) k_max <- n
  k_max <- min(k_max, n)

  # tent function values: tent[i, j] = Lambda_{(b_i, d_i)}(t_seq[j])
  tent <- matrix(0, nrow = n, ncol = resolution)
  for (i in seq_len(n)) {
    tent[i, ] <- pmax(0, pmin(t_seq - b[i], d[i] - t_seq))
  }

  # landscape[k, j] = kth largest tent value at t_seq[j]
  landscape <- matrix(0, nrow = k_max, ncol = resolution)
  for (j in seq_len(resolution)) {
    landscape[, j] <- sort(tent[, j], decreasing = TRUE)[seq_len(k_max)]
  }

  data.frame(
    t = rep(t_seq, each = k_max),
    k = rep(seq_len(k_max), times = resolution),
    value = as.vector(landscape)
  )
}

#' Compute the boundary operator for a simplicial complex
#'
#' @param simplices A list of simplices (each a numeric vector).
#' @param bound_dim The dimension k of the boundary operator \eqn{\partial_{k}}.
#' @return A sparse matrix representing \eqn{\partial_{k}}.
#'
#' @details
#' \deqn{\partial_k \sigma = \sum_i (-1)^i [v_0 v_1 \ldots \hat{v}_i \ldots v_k]}
#'
#' @importFrom Matrix sparseMatrix Matrix
#' @importFrom stats setNames
#' @export
#' @examples
#' simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))
#' boundary(simplices, 0)
boundary <- function(
    simplices, bound_dim
) {
  simplices_dim <- faces(simplices, target_dim=bound_dim)

  # No simplices at all in this dimension, return 0x0 matrix
  if (length(simplices_dim) == 0) {
    message(sprintf("No %d-simplices found, returning 0x0 matrix.", bound_dim))
    return(sparseMatrix(i = integer(0), j = integer(0), dims = c(0, 0)))
  }

  # No lower-dimensional simplices, return identity matrix
  if (bound_dim == 0) {
    n <- length(simplices_dim)
    message(sprintf("No %d-dimensional lower simplices, returning identity matrix.", bound_dim - 1))
    i <- rep(1, n) # row
    j <- 1:n # column
    x <- rep(1, n) # value
    Sparse <- sparseMatrix(i=i, j=j, x=x)

  }
  else {
    simplices_lower_dim <- faces(simplices, target_dim = (bound_dim - 1))

    # add index to simplices
    keys_dim <- sapply(simplices_dim, function(x) paste(sort(x), collapse= "-"))
    dct_dim <- setNames(seq_along(simplices_dim), keys_dim)

    keys_lower_dim <- sapply(simplices_lower_dim, function(x) paste(sort(x), collapse= "-"))
    dct_lower_dim <- setNames(seq_along(simplices_lower_dim), keys_lower_dim)

    Sparse <- Matrix(
      0,
      nrow = length(simplices_lower_dim),
      ncol = length(simplices_dim),
      sparse = TRUE
    )

    for (simplex_dim in simplices_dim) {
      j_key <- paste(sort(simplex_dim), collapse = "-")
      j <- dct_dim[j_key]

      i_entries <- list()
      for (face in seq_along(simplex_dim)) {
        simplex_lower <- simplex_dim[-face]
        i_key <- paste(sort(simplex_lower), collapse = "-")
        i <- dct_lower_dim[i_key]
        i_entries[[length(i_entries) + 1]] <- list(i = i, key = i_key)

        # Fill boundary matrix
        if (!is.na(i) && !is.na(j)) {
          Sparse[i, j] <- ifelse(face %% 2 == 0, -1, 1)
        }
      }

      # Print i and j only once per simplex
      i_message <- paste(sapply(i_entries, function(e) {
        sprintf("value: %d, key: %s", e$i, e$key)
      }), collapse = " I ")
      message("Lower dimension Simplices - ", i_message)
      message(sprintf("Simplices - value: %d, key: %s", j, j_key))
    }

  }
  return(Sparse)
}

source("./R/FloodComplex.R")

set.seed(42)
n_pts <- 200000
n_lms <- 250

generate_noisy_torus_points <- function(n, noise_sd = 0.05) {
  R <- 2; r <- 1
  u <- stats::runif(n, 0, 2 * pi)
  # rejection sampling for uniformity on the torus surface
  v <- numeric(n); k <- 0
  while (k < n) {
    cand <- stats::runif(n, 0, 2 * pi)
    acc <- stats::runif(n) < (R + r * cos(cand)) / (R + r)
    take <- min(n - k, sum(acc))
    if (take > 0) { v[(k + 1):(k + take)] <- cand[acc][seq_len(take)]; k <- k + take }
  }
  x <- (R + r * cos(v)) * cos(u)
  y <- (R + r * cos(v)) * sin(u)
  z <- r * sin(v)
  cbind(x, y, z) + matrix(stats::rnorm(3 * n, sd = noise_sd), ncol = 3)
}

pts <- generate_noisy_torus_points(n_pts)

t0 <- Sys.time()
filtration <- build_flood_filtration(pts, landmarks = n_lms, points_per_edge = 15, backend = "cpu")
t1 <- Sys.time()
cat(sprintf("Flood filtration: %d simplices in %.1fs\n", length(filtration), as.numeric(t1 - t0, units = "secs")))

pairs <- flood_persistence(filtration)
t2 <- Sys.time()
cat(sprintf("Persistence: %.1fs\n", as.numeric(t2 - t1, units = "secs")))

# most persistent intervals per dimension
pairs$pers <- pairs$death - pairs$birth
for (d in 0:2) {
  pd <- pairs[pairs$dim == d, ]
  pd <- pd[order(-pd$pers), ]
  cat(sprintf("\nH%d (top 5 of %d):\n", d, nrow(pd)))
  print(head(pd[, c("birth", "death", "pers")], 5), row.names = FALSE)
}

# persistence diagram
fin <- pairs[is.finite(pairs$death), ]
lim <- c(0, max(fin$death) * 1.05)
plot(fin$birth, fin$death, col = fin$dim + 1, pch = 19, cex = 0.7,
     xlim = lim, ylim = lim, xlab = "Birth", ylab = "Death",
     main = sprintf("Flood PH, noisy torus (%d pts, %d landmarks)", n_pts, n_lms))
abline(0, 1, lty = 2, col = "grey")
legend("bottomright", legend = paste0("H", 0:2), col = 1:3, pch = 19)

#' Construct a Vietoris–Rips Complex (1-skeleton + maximal simplices)
#'
#' @param points A numeric matrix or data.frame with one point per row (columns are coordinates).
#' @param epsilon A positive numeric threshold; connect points with distance < \eqn{\epsilon}.
#' @return A list with:
#' \describe{
#'   \item{network}{An \code{igraph} object representing the 1-skeleton.}
#'   \item{simplices}{A list of integer vectors, each the vertex indices of a maximal simplex.}
#' }
#'
#' @details
#' The Vietoris–Rips complex at scale \eqn{\epsilon} includes a simplex for every
#' finite set of points with pairwise distances < \eqn{\epsilon}. This function
#' constructs the 1-skeleton (edges only) and then uses maximal cliques in that
#' graph as the maximal simplices.
#'
#' @importFrom igraph graph.empty add_edges max_cliques
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' epsilon <- 1.5
#' vr_complex <- VietorisRipsComplex(points, epsilon)
VietorisRipsComplex <- function(points, epsilon) {

  labels <- 1:nrow(points)

  # Init graph (k-skeleton)
  network <- graph.empty(n = length(labels), directed = FALSE)

  for (i in 1:(length(labels) - 1)) {
    for (j in (i + 1):length(labels)) {
      dist <- sqrt(sum((points[i, ] - points[j, ])^2))
      if (dist <= epsilon) {
        # Add edge
        network <- add_edges(network, c(i, j))
      }
    }
  }
  cliques <- max_cliques(network)
  simplices <- lapply(cliques, function(clique) {
    simplex <- sort(clique)
    return(as.vector(simplex))
  })

  return(list(network = network, simplices = simplices))
}

#' Plot Persistence Diagram
#'
#' @param df Dataframe from plot_persistence.
#' @return A ggplot2 object representing the persistence diagram.
#'
#' @importFrom ggplot2 ggplot aes geom_point geom_abline theme_minimal labs scale_x_continuous scale_y_continuous coord_fixed theme .data
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
#' res <- boundary_info(filtration)
#' pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
#' plot_persistence(pairs)
plot_persistence <- function(df) {
  # add liitle padding for infinite death or it will cut the point
  finite_deaths <- df$death[is.finite(df$death)]
  death_max <- max(finite_deaths, na.rm = TRUE)
  death_pad <- 0.2 * death_max
  death_plot <- ifelse(is.infinite(df$death), death_max + death_pad, df$death)
  birth_min <- min(df$birth, na.rm = TRUE)
  x_lower <- min(0, birth_min - 0.05 * death_max)

  # "H0"/"H1"/"H2"... as a discrete factor, not the raw integer dim column,
  # so each dimension gets its own qualitative colour and legend entry
  # instead of shading indistinguishably along a continuous gradient.
  df$dim_label <- factor(paste0("H", df$dim), levels = paste0("H", sort(unique(df$dim))))

  plot <- ggplot(df, aes(x = .data$birth, y = death_plot, color = .data$dim_label)) +
    geom_point(size = 3, alpha = 0.7) +
    geom_abline(slope = 1, intercept = 0, linetype = "dashed") +
    theme_minimal() +
    labs(title = "Persistence Diagram", x = "Birth", y = "Death", color = "Dimension") +
    scale_x_continuous(limits = c(x_lower, death_max + death_pad), expand = c(0,0)) +
    scale_y_continuous(limits = c(0, death_max + death_pad + 0.1), expand = c(0,0)) +
    coord_fixed() + # same x, y scale
    theme(legend.position = "right")

  return(plot)
}

#' Plot a Persistence Landscape
#'
#' @param landscape_df Data frame from \code{persistence_landscape()}, with
#'   columns \code{t}, \code{k}, \code{value}.
#' @return A ggplot2 object with one line per landscape level
#'   \eqn{\lambda(k, \cdot)}.
#'
#' @importFrom ggplot2 ggplot aes geom_line theme_minimal labs theme .data
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
#' res <- boundary_info(filtration)
#' pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
#' landscape <- persistence_landscape(pairs, dimension = 0)
#' plot_landscape(landscape)
plot_landscape <- function(landscape_df) {
  plot <- ggplot(landscape_df, aes(x = .data$t, y = .data$value,
                                    color = factor(.data$k), group = .data$k)) +
    geom_line() +
    theme_minimal() +
    labs(title = "Persistence Landscape", x = "t", y = "Landscape value", color = "k") +
    theme(legend.position = "right")

  return(plot)
}

#' Build a cubical (grid) filtration from an image
#'
#' @param image A numeric matrix (grid of pixel/voxel values, e.g. a
#'   grayscale image with values in \code{[0, 255]}).
#' @param superlevel If \code{TRUE}, filters by decreasing value instead
#'   (equivalent to negating \code{image} first) - useful for tracking
#'   bright structures shrinking rather than dark structures growing.
#' @return A filtration list: one \code{list(simplex = <integer vector of
#'   pixel ids>, t = <numeric>)} per simplex, sorted by (t, dimension,
#'   lexicographic order) - the same format as
#'   \code{\link{build_filtration}}.
#'
#' @export
#' @examples
#' # a ring (value 1) around a hole (value 5) around a background (value 9):
#' # the hole is born once the ring closes and dies once its center fills in
#' image <- matrix(c(
#'   9, 9, 9, 9, 9,
#'   9, 1, 1, 1, 9,
#'   9, 1, 5, 1, 9,
#'   9, 1, 1, 1, 9,
#'   9, 9, 9, 9, 9
#' ), nrow = 5, byrow = TRUE)
#' filtration <- build_cubical_filtration(image)
#' pairs <- persistence_pairs(filtration)
#' pairs[pairs$dim == 1, ] # one H1 bar: birth = 1 (ring closes), death = 5
#' plot_persistence(pairs)
build_cubical_filtration <- function(image, superlevel = FALSE) {
  image <- as.matrix(image)
  nr <- nrow(image)
  nc <- ncol(image)

  value <- if (superlevel) -image else image

  vid <- function(i, j) (i - 1L) * nc + j # 1-based linear pixel id, row-major
  coord <- function(v) {
    v0 <- v - 1L
    c(v0 %/% nc + 1L, v0 %% nc + 1L) # c(row, col)
  }
  cell_value <- function(verts) {
    rc <- vapply(verts, coord, numeric(2))
    max(value[cbind(rc[1, ], rc[2, ])])
  }

  simplex_info <- vector("list", 5L * nr * nc)
  idx <- 0L
  for (i in seq_len(nr)) {
    for (j in seq_len(nc)) {
      idx <- idx + 1L
      simplex_info[[idx]] <- list(simplex = vid(i, j), t = value[i, j])
    }
  }

  seen <- new.env(hash = TRUE, parent = emptyenv())

  add_simplex <- function(verts) {
    key <- paste(sort(verts), collapse = "-")
    if (exists(key, envir = seen, inherits = FALSE)) return(invisible(NULL))
    assign(key, TRUE, envir = seen)
    idx <<- idx + 1L
    simplex_info[[idx]] <<- list(simplex = sort(verts), t = cell_value(verts))
  }

  for (i in seq_len(nr - 1L)) {
    for (j in seq_len(nc - 1L)) {
      v00 <- vid(i, j);      v10 <- vid(i + 1L, j)
      v01 <- vid(i, j + 1L); v11 <- vid(i + 1L, j + 1L)

      add_simplex(c(v00, v10)) # left edge
      add_simplex(c(v00, v01)) # top edge
      add_simplex(c(v10, v11)) # right edge
      add_simplex(c(v01, v11)) # bottom edge
      add_simplex(c(v00, v11)) # diagonal

      add_simplex(c(v00, v10, v11))
      add_simplex(c(v00, v01, v11))
    }
  }

  simplex_info <- simplex_info[seq_len(idx)]

  ord <- order(
    vapply(simplex_info, `[[`, 0, "t"),
    lengths(lapply(simplex_info, `[[`, "simplex")),
    vapply(simplex_info, function(x) paste(x$simplex, collapse = "-"), "")
  )
  simplex_info[ord]
}

#' Generate all unique faces of a given dimension from simplices
#'
#' @param simplices A list of simplices (each a numeric vector).
#' @param target_dim The target dimension \eqn{k} for the faces (e.g., \code{0} for vertices, \code{1} for edges, etc.).
#' @return A list of faces (each a numeric vector) of dimension \code{target_dim}.
#'
#' @details
#' The function generates all possible subsets (combinations) of each simplex, removes duplicates,
#' and filters them to only include those of length \code{target_dim + 1}.
#'
#' For example, a 2-simplex \code{c(1, 2, 3)} has three 1-dimensional faces (edges):
#' \code{c(1,2)}, \code{c(1,3)}, and \code{c(2,3)}, and three 0-dimensional faces (vertices):
#' \code{1}, \code{2}, and \code{3}.
#'
#' @importFrom gtools combinations
#' @export
#' @examples
#' simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))
#' faces(simplices, target_dim=0)
faces <- function(
    simplices, target_dim
) {

  faceset <- list()
  facestr_set <- character()

  target_len <- target_dim + 1

  for (simplex in simplices) {

    node_counts <- length(simplex)

    if (node_counts < target_len) {next}

    combs <- combinations(n = node_counts, r = target_len, v = simplex)

    for (i in 1:nrow(combs)) {
      face <- sort(combs[i, ])
      face_str <- paste(face, collapse = "-")

      if (!(face_str %in% facestr_set)) {
        faceset[[length(faceset) + 1]] <- face
        facestr_set <- c(facestr_set, face_str)
      }
    }
  }

  return(faceset)
}

#' Build a filtration from a point cloud, for any of five complex types
#'
#' @param points A numeric matrix or data.frame, one point per row.
#' @param method One of \code{"VR"}, \code{"Delaunay"}, \code{"Alpha"},
#'   \code{"Cech"}, \code{"Witness"}.
#' @param eps_max Maximum scale. Required for \code{"VR"}, \code{"Cech"} and
#'   \code{"Witness"}; for \code{"Alpha"} it caps the alpha value (defaults
#'   to \code{Inf}, i.e. no cap); ignored for \code{"Delaunay"} (which has no
#'   scale parameter - see \code{\link{DelaunayComplex}}).
#' @param landmarks For \code{method = "Witness"} only: either an integer
#'   (number of landmarks to select via \code{\link{generate_landmarks}}) or
#'   an integer vector of landmark row indices into \code{points}. Required
#'   for \code{"Witness"}.
#' @param nu For \code{method = "Witness"} only: the landmark-distance
#'   parameter passed to \code{\link{WitnessComplex}}. Defaults to \code{1}.
#' @param max_dimension Optional maximum homology dimension you intend to
#'   compute persistence for.
#' @return A filtration list, as described above. When \code{max_dimension}
#'   is set, the list carries it as a \code{"max_dimension"} attribute (see
#'   above); otherwise the attribute is absent, exactly as before this
#'   parameter existed.
#'
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' vr_filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
#' alpha_filtration <- build_filtration(points, method = "Alpha")
#' delaunay_filtration <- build_filtration(points, method = "Delaunay")
#' cech_filtration <- build_filtration(points, method = "Cech", eps_max = 0.8)
#' # cap at H1: persistence_pairs()/etc. need no max_dimension of their own
#' vr_capped <- build_filtration(points, method = "VR", eps_max = 1.2,
#'                                max_dimension = 1)
#' pairs <- persistence_pairs(vr_capped)
build_filtration <- function(points, method = c("VR", "Delaunay", "Alpha", "Cech", "Witness"),
                              eps_max = NULL, landmarks = NULL, nu = 1, max_dimension = NULL) {
  method <- match.arg(method)
  points <- as.matrix(points)

  if (method %in% c("VR", "Cech", "Witness") && is.null(eps_max)) {
    stop(sprintf("method = '%s' requires eps_max.", method))
  }
  if (method == "Witness" && is.null(landmarks)) {
    stop("method = 'Witness' requires landmarks (a count or a vector of row indices).")
  }

  # The "+1 trick": one extra dimension kept internally so dimension-
  # max_dimension classes are still correctly killed - see
  # restrict_filtration()'s Details. NULL (uncapped) stays NULL.
  build_kmax <- if (is.null(max_dimension)) NULL else max_dimension + 1

  result <- switch(
    method,
    VR = {
      vr <- VietorisRipsComplex(points, epsilon = eps_max)
      simplices_to_filtration(vr$simplices, function(s) vr_scale_of_simplex(points, s),
                               max_dimension = build_kmax)
    },
    Cech = {
      cc <- CechComplex(points, epsilon = eps_max)
      simplices_to_filtration(cc$simplices, function(s) cech_scale_of_simplex(points, s),
                               max_dimension = build_kmax)
    },
    Witness = {
      wc <- WitnessComplex(points, landmarks = landmarks, epsilon = eps_max, nu = nu)
      simplices_to_filtration(wc$simplices, function(s) witness_scale_of_simplex(wc$edge_birth, s),
                               max_dimension = build_kmax)
    },
    Alpha = {
      epsilon <- if (is.null(eps_max)) Inf else eps_max
      alpha_filtration <- as_filtration(AlphaComplex(points, epsilon = epsilon))
      # AlphaComplex() has no build-time dimension cap, so restrict after
      # the fact instead - still correct (just not cheaper to build).
      if (is.null(build_kmax)) alpha_filtration else restrict_filtration(alpha_filtration, build_kmax)
    },
    Delaunay = {
      delaunay_filtration <- as_filtration(DelaunayComplex(points))
      if (is.null(build_kmax)) delaunay_filtration else restrict_filtration(delaunay_filtration, build_kmax)
    }
  )

  # restrict_filtration() filters a plain list (via Filter()/subsetting),
  # which drops attributes - so max_dimension must be tagged here, after
  # the switch, not earlier. NULL is a safe no-op when no cap was requested.
  attr(result, "max_dimension") <- max_dimension
  result
}

#' @keywords internal
vr_scale_of_simplex <- function(points, simplex_idx) {
  # Get the maximum distance between points in the simplex
  if (length(simplex_idx) <= 1) return(0)
  # Get specific rows
  srow <- points[simplex_idx, , drop=FALSE]
  dmax <- 0
  # Loop will give all the conditions
  # If nrow = 4, 1 -> 2,3,4. 2 -> 3,4. 3 -> 4
  for (i in 1:(nrow(srow)-1)) {
    for (j in (i+1):nrow(srow)) {
      d <- sqrt(sum((srow[i,]-srow[j,])^2))
      if (d > dmax) dmax <- d
    }
  }
  return(dmax)
}

# Construct a Alpha complex
#
# S is a finite set of points in R^d. Using the Voronoi decomposition
# {V_s}_{s in S}, the alpha complex A_epsilon(S) is the nerve of the cover
# {V_s ∩ B(s, epsilon)}_{s in S} of S_epsilon = union_s B(s, epsilon); as
# epsilon -> Inf this cover tends to the Voronoi decomposition itself, and
# A_Inf(S) is exactly the Delaunay complex. Rather than computing that nerve directly, we use the
# standard equivalent computational recipe "Three-Dimensional Alpha Shapes"

#' @keywords internal
.alpha_values <- function(points, top_simplices) {
  points <- as.matrix(points)
  n <- nrow(points)
  d_top <- max(sapply(top_simplices, length)) - 1

  faces_by_dim <- lapply(0:d_top, function(k) faces(top_simplices, k))
  key_of <- function(s) paste(sort(s), collapse = "-")

  filt_val <- new.env(hash = TRUE, parent = emptyenv())

  # top dimension: always Gabriel by the Delaunay empty-circumsphere property
  for (s in faces_by_dim[[d_top + 1]]) {
    cs <- circumsphere(points[s, , drop = FALSE])
    assign(key_of(s), cs$radius, envir = filt_val)
  }

  if (d_top >= 1) {
    for (k in (d_top - 1):0) {
      cur_faces <- faces_by_dim[[k + 1]]
      higher_faces <- faces_by_dim[[k + 2]]

      # push each already-timed (k+1)-face's value down onto its own
      # k-dim sub-faces: a non-Gabriel sigma inherits min(alpha(tau)) over
      # its cofaces tau, and every coface has by now been timed.
      cand <- new.env(hash = TRUE, parent = emptyenv())
      for (t in higher_faces) {
        tval <- get(key_of(t), envir = filt_val)
        subfaces <- if (length(t) == 2) as.list(t) else {
          m <- utils::combn(t, length(t) - 1)
          lapply(seq_len(ncol(m)), function(j) m[, j])
        }
        for (sf in subfaces) {
          skey <- key_of(sf)
          prev <- mget(skey, envir = cand, ifnotfound = list(Inf))[[1]]
          if (tval < prev) assign(skey, tval, envir = cand)
        }
      }

      for (s in cur_faces) {
        cs <- circumsphere(points[s, , drop = FALSE])
        others <- setdiff(seq_len(n), s)
        gabriel <- TRUE
        if (length(others) > 0) {
          dd <- sqrt(rowSums(sweep(points[others, , drop = FALSE], 2, cs$center)^2))
          tol <- 1e-8 * max(1, cs$radius)
          if (any(dd < cs$radius - tol)) gabriel <- FALSE
        }
        val <- if (gabriel) cs$radius else get(key_of(s), envir = cand)
        assign(key_of(s), val, envir = filt_val)
      }
    }
  }

  list(faces_by_dim = faces_by_dim, filt_val = filt_val, key_of = key_of)
}

#' Construct an Alpha Complex
#'
#' @param points A numeric matrix or data.frame with one point per row
#'   (columns are coordinates); must be in "general position" (see
#'   \code{\link{circumsphere}}).
#' @param epsilon The alpha-complex scale. A simplex is included once its
#'   alpha value is at most \code{epsilon}. Defaults to \code{Inf}, which
#'   gives the full Delaunay complex (see \code{\link{DelaunayComplex}}).
#' @return A list of class \code{"alpha_complex"} with elements:
#' \describe{
#'   \item{simplices}{A list of integer vectors, every included simplex (all
#'     dimensions, not just the maximal ones - membership in the alpha
#'     complex is not simply generated by cliques).}
#'   \item{filtration}{A numeric vector, the alpha value of each simplex in
#'     \code{simplices}, in the same order.}
#' }
#' Pass the result to \code{\link{as_filtration}} for the same
#' \code{list(simplex=, t=)} format used by \code{\link{build_filtration}}
#' and every persistence function, or just call
#' \code{build_filtration(points, method = "Alpha", eps_max = epsilon)}.
#'
#' @importFrom geometry delaunayn
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' alpha_complex <- AlphaComplex(points, epsilon = 1)
AlphaComplex <- function(points, epsilon = Inf) {
  points <- as.matrix(points)
  delaunay <- geometry::delaunayn(points, options = "Qt Qbb Qc Qz")
  top_simplices <- lapply(seq_len(nrow(delaunay)), function(i) sort(delaunay[i, ]))

  av <- .alpha_values(points, top_simplices)
  all_faces <- unlist(av$faces_by_dim, recursive = FALSE)
  values <- vapply(all_faces, function(s) get(av$key_of(s), envir = av$filt_val), 0)

  keep <- values <= epsilon
  structure(
    list(simplices = all_faces[keep], filtration = values[keep]),
    class = "alpha_complex"
  )
}

#' Construct a Delaunay Complex
#'
#' @param points A numeric matrix or data.frame with one point per row
#'   (columns are coordinates); must be in "general position".
#' @return A list of class \code{c("delaunay_complex", "alpha_complex")};
#'   see \code{\link{AlphaComplex}} for the element description.
#'
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' delaunay_complex <- DelaunayComplex(points)
DelaunayComplex <- function(points) {
  result <- AlphaComplex(points, epsilon = Inf)
  class(result) <- c("delaunay_complex", "alpha_complex")
  result
}

#' Build the clique (flag) complex of a graph on a fixed vertex set
#'
#' Shared by \code{\link{VietorisRipsComplex}}, \code{\link{CechComplex}} and
#' \code{\link{WitnessComplex}}: each of them reduces to "connect vertices
#' that are close enough (in whatever sense that complex uses), then take the
#' maximal cliques of that graph as the maximal simplices."
#'
#' @param n Number of vertices.
#' @param edges An integer matrix with two columns (or a length-\code{2k}
#'   vector, igraph style), one row per edge, 1-based vertex indices.
#' @return A list with \code{network} (an \code{igraph} object, the
#'   1-skeleton) and \code{simplices} (a list of integer vectors, the
#'   vertex sets of the maximal cliques, each sorted).
#'
#' @importFrom igraph graph.empty add_edges max_cliques
#' @keywords internal
build_clique_complex <- function(n, edges) {
  network <- graph.empty(n = n, directed = FALSE)
  if (length(edges) > 0) {
    edges_mat <- matrix(edges, ncol = 2)
    network <- add_edges(network, as.vector(t(edges_mat)))
  }
  cliques <- max_cliques(network)
  simplices <- lapply(cliques, function(clique) as.vector(sort(clique)))
  list(network = network, simplices = simplices)
}

#' Expand maximal simplices into a sorted filtration list
#'
#' Shared filtration-assembly step used by every \code{build_filtration()}
#' method: take a set of maximal simplices, generate every face of every
#' dimension via \code{\link{faces}}, assign each face a filtration time via
#' \code{scale_fn}, and sort by (time, dimension, lexicographic order) - the
#' same convention used throughout the package (see \code{\link{faces}} for
#' what "lexicographic order" means here).
#'
#' @param maximal_simplices A list of integer vectors (the maximal simplices).
#' @param scale_fn A function taking one simplex (integer vector) and
#'   returning its filtration time.
#' @param max_dimension Optional integer cap. If supplied, faces of dimension
#'   greater than \code{max_dimension} are never generated in the first
#'   place - this is a structural cap on \code{kmax}, evaluated BEFORE
#'   \code{\link{faces}} is called, not a post-hoc filter. Callers that need
#'   the persistence "+1 trick" (see \code{\link{restrict_filtration}}) are
#'   responsible for passing \code{max_dimension + 1} here, not
#'   \code{max_dimension} itself - this function does not know about that
#'   convention.
#' @return A filtration list: one \code{list(simplex = <integer vector>, t =
#'   <numeric>)} per face, sorted by (t, dimension, lexicographic order).
#'
#' @keywords internal
simplices_to_filtration <- function(maximal_simplices, scale_fn, max_dimension = NULL) {
  kmax <- max(sapply(maximal_simplices, length)) - 1
  if (!is.null(max_dimension)) {
    kmax <- min(kmax, max_dimension)
  }
  face <- lapply(0:kmax, function(k) faces(maximal_simplices, k))
  all_faces <- unlist(face, recursive = FALSE)

  simplex_info <- lapply(all_faces, function(s) list(simplex = s, t = scale_fn(s)))
  ord <- order(sapply(simplex_info, function(x) x[["t"]]),
               sapply(simplex_info, function(x) length(x$simplex)),
               sapply(simplex_info, function(x) paste(x$simplex, collapse = "-")))

  simplex_info[ord]
}

#' Pairwise Euclidean distance matrix
#'
#' @param points A numeric matrix, one point per row.
#' @param query Optional second numeric matrix; if supplied, returns the
#'   \code{nrow(points)} x \code{nrow(query)} cross-distance matrix instead of
#'   the full pairwise matrix of \code{points} with itself.
#' @return A numeric distance matrix.
#'
#' @keywords internal
pairwise_dist <- function(points, query = NULL) {
  a <- as.matrix(points)
  b <- if (is.null(query)) a else as.matrix(query)
  aa <- rowSums(a^2)
  bb <- rowSums(b^2)
  d2 <- outer(aa, bb, "+") - 2 * (a %*% t(b))
  sqrt(pmax(d2, 0))
}

#' Minimum enclosing ball of a finite point set (Welzl's algorithm)
#'
#' Used by \code{\link{CechComplex}}: the Cech complex includes a simplex
#' \eqn{\sigma} at scale \eqn{\epsilon} exactly when the balls of radius
#' \eqn{\epsilon} centered at its vertices have a common point, which happens
#' if and only if the minimum enclosing ball of \eqn{\sigma}'s vertices has
#' radius at most \eqn{\epsilon} (this is the standard reduction used e.g. by
#' GUDHI's Cech complex; see Cavanna, Jahanseir and Sheehy (2017)).
#'
#' Randomized, expected linear-time in the number of points; the point sets
#' passed in here are the vertices of a single simplex, so they are always
#' small.
#'
#' @param points A numeric matrix, one point per row (at least 1 row).
#' @return A list with \code{center} (numeric vector) and \code{radius}.
#'
#' @keywords internal
min_enclosing_ball <- function(points) {
  points <- as.matrix(points)
  d <- ncol(points)

  ball_of <- function(boundary) {
    # boundary: matrix of <= d+1 points that must lie ON the returned ball.
    m <- nrow(boundary)
    if (m == 0) return(list(center = rep(NA_real_, d), radius = -Inf))
    if (m == 1) return(list(center = boundary[1, ], radius = 0))
    b1 <- boundary[1, ]
    V <- sweep(boundary[-1, , drop = FALSE], 2, b1, "-")
    G <- V %*% t(V)
    q <- 0.5 * rowSums(V^2)
    lambda <- tryCatch(solve(G, q), error = function(e) {
      stop("Points are not in general position: cannot determine an enclosing ball.")
    })
    center <- b1 + as.vector(lambda %*% V)
    list(center = center, radius = sqrt(sum((center - b1)^2)))
  }

  in_ball <- function(p, ball, tol = 1e-9 * max(1, ball$radius)) {
    sqrt(sum((p - ball$center)^2)) <= ball$radius + tol
  }

  welzl <- function(pts, boundary) {
    if (nrow(pts) == 0 || nrow(boundary) == d + 1) {
      return(ball_of(boundary))
    }
    p <- pts[1, ]
    rest <- pts[-1, , drop = FALSE]
    ball <- welzl(rest, boundary)
    if (is.finite(ball$radius) && in_ball(p, ball)) return(ball)
    welzl(rest, rbind(boundary, p))
  }

  # shuffle for the algorithm's expected linear-time guarantee; use a fixed
  # (non-random) rotation instead of sample() so results are reproducible
  # without disturbing the caller's RNG state.
  n <- nrow(points)
  ord <- if (n <= 1) seq_len(n) else c(2:n, 1)
  welzl(points[ord, , drop = FALSE], points[0, , drop = FALSE])
}

#' Circumsphere of an affinely independent point set
#'
#' The unique sphere through \code{points} whose center lies in their affine
#' hull - i.e. the minimal-radius sphere with all of \code{points} on its
#' boundary. Used by \code{\link{AlphaComplex}}/\code{\link{DelaunayComplex}}
#' to compute alpha values (unlike \code{\link{min_enclosing_ball}}, which
#' minimizes radius over ALL enclosing balls, this fixes every point to lie
#' exactly on the sphere - the two coincide only for simplices whose
#' circumcenter is not "hidden" inside the simplex in a way that admits a
#' smaller covering ball).
#'
#' @param points A numeric matrix, one point per row (must be affinely
#'   independent, i.e. "general position" - see Otter et al. (2017)).
#' @return A list with \code{center} (numeric vector) and \code{radius}.
#'
#' @keywords internal
circumsphere <- function(points) {
  points <- as.matrix(points)
  m <- nrow(points)
  if (m == 1) return(list(center = points[1, ], radius = 0))

  b1 <- points[1, ]
  V <- sweep(points[-1, , drop = FALSE], 2, b1, "-")
  G <- V %*% t(V)
  q <- 0.5 * rowSums(V^2)
  lambda <- tryCatch(solve(G, q), error = function(e) {
    stop("Points are not in general position: cannot compute a circumsphere.")
  })
  center <- b1 + as.vector(lambda %*% V)
  list(center = center, radius = sqrt(sum((center - b1)^2)))
}

#' Wasserstein/bottleneck distance between two point clouds, via one chosen complex
#'
#' Picks a single \code{\link{build_filtration}} method, builds the
#' persistence diagram of each point cloud with it, and compares the two
#' diagrams with either \code{\link{wasserstein_distance}} or
#' \code{\link{bottleneck_distance}}. Useful for asking "how different are
#' these two datasets topologically", holding the complex construction fixed
#' so the comparison is apples-to-apples.
#'
#' @param points1,points2 Numeric matrices or data.frames, one point per row.
#' @param method One of \code{"VR"}, \code{"Delaunay"}, \code{"Alpha"},
#'   \code{"Cech"}, \code{"Witness"} - passed to \code{\link{build_filtration}}.
#' @param distance One of \code{"wasserstein"} or \code{"bottleneck"}.
#' @param dimension Homological dimension to compare.
#' @param p Wasserstein order, used only when \code{distance = "wasserstein"}.
#' @param ground Ground metric passed to the chosen distance function
#'   (\code{"L2"} or \code{"Linf"}). Defaults to each distance's own default
#'   (\code{"L2"} for Wasserstein, \code{"Linf"} for bottleneck) if left
#'   \code{NULL}.
#' @param eps_max Maximum scale, passed to \code{\link{build_filtration}}.
#'   Required for \code{"VR"}, \code{"Cech"} and \code{"Witness"}; optional
#'   for \code{"Alpha"} (defaults to \code{Inf}); ignored for
#'   \code{"Delaunay"}.
#' @param landmarks Passed to \code{\link{build_filtration}} for
#'   \code{"Witness"}; if \code{NULL}, defaults independently for each point
#'   cloud to \code{min(30, nrow(points))} landmarks, with a message.
#' @param nu Passed to \code{\link{build_filtration}} for \code{"Witness"}.
#' @return A list with:
#' \describe{
#'   \item{distance}{The computed distance (a single number).}
#'   \item{method, distance_type, dimension}{Echoed back for reference.}
#'   \item{diagram1, diagram2}{The two persistence diagrams (data frames)
#'     the distance was computed from.}
#' }
#'
#' @export
#' @examples
#' set.seed(1)
#' cloud_a <- matrix(rnorm(24), ncol = 2)
#' cloud_b <- matrix(rnorm(24), ncol = 2) + 0.1
#' complex_distance(cloud_a, cloud_b, method = "VR", distance = "bottleneck",
#'                   dimension = 0, eps_max = 0.6)
complex_distance <- function(points1, points2,
                              method = c("VR", "Delaunay", "Alpha", "Cech", "Witness"),
                              distance = c("wasserstein", "bottleneck"),
                              dimension = 0,
                              p = 2,
                              ground = NULL,
                              eps_max = NULL,
                              landmarks = NULL,
                              nu = 1) {
  method <- match.arg(method)
  distance <- match.arg(distance)
  points1 <- as.matrix(points1)
  points2 <- as.matrix(points2)

  if (method %in% c("VR", "Cech", "Witness") && is.null(eps_max)) {
    stop(sprintf("eps_max is required for method = '%s'.", method))
  }

  build_one <- function(points, lm) {
    if (method == "Delaunay") {
      build_filtration(points, method = method)
    } else if (method == "Witness") {
      build_filtration(points, method = method, eps_max = eps_max, landmarks = lm, nu = nu)
    } else {
      build_filtration(points, method = method, eps_max = eps_max)
    }
  }

  lm1 <- landmarks
  lm2 <- landmarks
  if (method == "Witness" && is.null(landmarks)) {
    lm1 <- min(30L, nrow(points1))
    lm2 <- min(30L, nrow(points2))
    message(sprintf(
      "No landmarks supplied for method = 'Witness'; using %d and %d landmarks (points1, points2 respectively).",
      lm1, lm2
    ))
  }

  diagram1 <- persistence_pairs(build_one(points1, lm1))
  diagram2 <- persistence_pairs(build_one(points2, lm2))

  if (is.null(ground)) ground <- if (distance == "bottleneck") "Linf" else "L2"

  value <- if (distance == "wasserstein") {
    wasserstein_distance(diagram1, diagram2, dimension = dimension, p = p, ground = ground)
  } else {
    bottleneck_distance(diagram1, diagram2, dimension = dimension, ground = ground)
  }

  list(distance = value, method = method, distance_type = distance, dimension = dimension,
       diagram1 = diagram1, diagram2 = diagram2)
}

# ---------------------------------------------------------------------------
# Flood Complex for the SimplicialComplex package
#
# Pure-R re-implementation of the "flooder" Python package:
#   Graf, Pellizzoni, Uray, Huber, Kwitt (NeurIPS 2025),
#   "The Flood Complex: Large-Scale Persistent Homology on Millions of Points"
#   https://github.com/plus-rkwitt/flooder  (MIT)
#
# Pipeline:
#   1. Select landmarks via Farthest-Point Sampling (FPS).
#   2. Build the Delaunay triangulation on the landmarks (geometry::delaunayn).
#   3. For every top-dimensional Delaunay simplex, place a barycentric grid of
#      sample points on it. The filtration value ("flood time") of a simplex is
#      the covering radius: the maximum, over its grid points, of the distance
#      to the nearest witness point. Faces inherit their values from the grid
#      points that lie on them.
#   4. Make the filtration non-decreasing, then compute persistent homology.
#
# Dependencies:
#   geometry (Delaunay), RANN (kd-tree NN queries)
#   Optional: torch (GPU acceleration of the distance computation)
# ---------------------------------------------------------------------------

#' Farthest-Point Sampling of landmarks
#'
#' Selects \code{n_lms} landmarks from a point cloud via (exact) Farthest-Point
#' Sampling. Equivalent to \code{flooder::generate_landmarks} (which uses an
#' approximate bucket-FPS; the exact version below gives the same qualitative
#' coverage).
#'
#' @param points A numeric matrix (N x d) point cloud.
#' @param n_lms Number of landmarks to sample (<= N).
#' @param start_idx Index of the starting point. Defaults to 1
#'   (flooder defaults to index 0, i.e. the same first point).
#' @return A list with \code{landmarks} (n_lms x d matrix) and
#'   \code{indices} (row indices into \code{points}).
#'
#' @export
#' @examples
#' \dontrun{
#' pts <- matrix(rnorm(2000), ncol = 2)
#' lms <- generate_landmarks(pts, 50)
#' }
generate_landmarks <- function(points, n_lms, start_idx = 1) {
  points <- as.matrix(points)
  n <- nrow(points)
  if (n_lms <= 0) stop("Number of landmarks must be positive")
  n_lms <- min(n_lms, n)

  idx <- integer(n_lms)
  idx[1] <- start_idx
  # squared distance of every point to the current landmark set
  d2 <- rowSums(sweep(points, 2, points[start_idx, ])^2)
  if (n_lms > 1) {
    for (i in 2:n_lms) {
      idx[i] <- which.max(d2)
      d2_new <- rowSums(sweep(points, 2, points[idx[i], ])^2)
      d2 <- pmin(d2, d2_new)
    }
  }
  list(landmarks = points[idx, , drop = FALSE], indices = idx)
}

#' @keywords internal
#' Barycentric grid on the unit simplex (replicates flooder's generate_grid).
#' Returns the weight matrix (C x (dim+1)) plus, for every face of the
#' standard simplex, which local vertices span it and which grid rows lie on it.
.flood_grid <- function(n, dim) {
  # all non-negative integer vectors of length dim+1 summing to n-1
  if (dim == 0) {
    grid <- matrix(n - 1, 1, 1)
  } else {
    combs <- t(utils::combn(0:(n + dim - 2), dim)) # C x dim
    padded <- cbind(-1, combs, n + dim - 1) # C x (dim+2)
    grid <- matrix(t(apply(padded, 1, diff)) - 1, ncol = dim + 1) # C x (dim+1)
  }
  weights <- grid / (n - 1)

  face_rows <- list() # rows of the grid lying on each face
  face_vertices <- list() # local vertex indices (1-based) spanning each face
  m <- 0
  for (k in 0:dim) { # k = number of barycentric coordinates forced to zero
    zero_sets <- utils::combn(seq_len(dim + 1), k, simplify = FALSE)
    if (k == 0) zero_sets <- list(integer(0))
    for (zs in zero_sets) {
      m <- m + 1
      if (length(zs) == 0) {
        rows <- seq_len(nrow(grid))
      } else {
        rows <- which(rowSums(grid[, zs, drop = FALSE] != 0) == 0)
      }
      face_rows[[m]] <- rows
      face_vertices[[m]] <- setdiff(seq_len(dim + 1), zs)
    }
  }
  list(weights = weights, face_rows = face_rows, face_vertices = face_vertices)
}

#' @keywords internal
#' Nearest-neighbour distances, CPU backend (kd-tree via RANN, with a
#' base-R brute-force fallback when RANN is not installed).
.flood_nn_cpu <- function(points, queries) {
  if (requireNamespace("RANN", quietly = TRUE)) {
    RANN::nn2(data = points, query = queries, k = 1)$nn.dists[, 1]
  } else {
    .flood_nn_base(points, queries)
  }
}

#' @keywords internal
#' Chunked brute-force nearest-neighbour distances in base R.
.flood_nn_base <- function(points, queries, chunk = 2048L) {
  pt2 <- rowSums(points^2)
  nq <- nrow(queries)
  out <- numeric(nq)
  for (s in seq(1L, nq, by = chunk)) {
    e <- min(nq, s + chunk - 1L)
    Q <- queries[s:e, , drop = FALSE]
    d2 <- outer(rowSums(Q^2), pt2, "+") - 2 * (Q %*% t(points))
    j <- max.col(-d2, ties.method = "first") # fast row-wise argmin
    out[s:e] <- sqrt(pmax(d2[cbind(seq_len(nrow(d2)), j)], 0))
  }
  out
}

#' @keywords internal
#' Nearest-neighbour distances, torch backend (GPU if CUDA is available).
#' Brute-force cdist in chunks; mirrors flooder's non-Triton CUDA path.
.flood_nn_torch <- function(points_t, queries, device, chunk = 4096L) {
  q <- torch::torch_tensor(queries, dtype = torch::torch_float32(), device = device)
  nq <- nrow(queries)
  out <- numeric(nq)
  for (s in seq(1L, nq, by = chunk)) {
    e <- min(nq, s + chunk - 1L)
    idx <- torch::torch_tensor(as.integer(s:e), dtype = torch::torch_long(),
                               device = device)
    d <- torch::torch_cdist(q$index_select(1L, idx), points_t)
    out[s:e] <- as.numeric(torch::torch_min(d, dim = 2L)[[1]]$cpu())
  }
  out
}

#' Construct a Flood complex
#'
#' Builds the Flood complex of a point cloud: a Delaunay complex on a set of
#' landmarks whose simplices are filtered by their covering radius with
#' respect to the full point cloud ("flood time").
#'
#' @param points A numeric matrix (N x d) of witness points.
#' @param landmarks Either an integer (number of FPS landmarks) or a numeric matrix (N_l x d) of explicit landmark coordinates.
#' @param max_dimension Top dimension of the simplices. Defaults to d.
#' @param points_per_edge Grid resolution per simplex edge (accuracy vs. speed trade-off). Defaults to 30, as in flooder.
#' @param backend One of \code{"auto"}, \code{"cpu"}, \code{"torch"}.
#'   \code{"cpu"} uses a kd-tree (RANN). \code{"torch"} uses the torch R package and runs on CUDA when available. \code{"auto"} picks torch only if a CUDA device is present, else the kd-tree.
#' @param batch_points Maximum number of grid points processed per batch (bounds memory). Defaults to 2^22.
#' @param delaunay Optional precomputed Delaunay triangulation of the landmarks: an m x (d+1) integer matrix of 1-based landmark indices. If NULL (default), computed via \code{geometry::delaunayn}.
#' @return A list of class \code{"flood_complex"} with elements
#'   \code{simplices} (list of integer vectors, landmark indices),
#'   \code{filtration} (numeric vector of flood times),
#'   \code{landmarks} (matrix), \code{landmark_indices} (or NULL).
#'
#' @export
#' @examples
#' \dontrun{
#' pts <- matrix(rnorm(3000), ncol = 2)
#' fc <- flood_complex(pts, landmarks = 40)
#' }
flood_complex <- function(points,
                          landmarks,
                          max_dimension = NULL,
                          points_per_edge = 30,
                          backend = c("auto", "cpu", "torch"),
                          batch_points = 2^22,
                          delaunay = NULL) {
  backend <- match.arg(backend)
  points <- as.matrix(points)
  d_amb <- ncol(points)
  if (is.null(max_dimension)) max_dimension <- d_amb

  lm_idx <- NULL
  if (length(landmarks) == 1 && is.numeric(landmarks)) {
    lm <- generate_landmarks(points, as.integer(landmarks))
    landmarks <- lm$landmarks
    lm_idx <- lm$indices
  } else {
    landmarks <- as.matrix(landmarks)
  }
  if (ncol(landmarks) != d_amb) stop("landmarks and points must have the same dimension")

  # backend selection
  use_torch <- FALSE
  device <- NULL
  if (backend == "torch" ||
      (backend == "auto" &&
       requireNamespace("torch", quietly = TRUE) &&
       torch::cuda_is_available())) {
    if (!requireNamespace("torch", quietly = TRUE)) {
      stop("backend = 'torch' requested but the torch package is not installed")
    }
    use_torch <- TRUE
    device <- if (torch::cuda_is_available()) "cuda" else "cpu"
    points_t <- torch::torch_tensor(points, dtype = torch::torch_float32(),
                                    device = device)
  }

  # Delaunay triangulation on the landmarks
  if (is.null(delaunay)) {
    if (!requireNamespace("geometry", quietly = TRUE)) {
      stop("Package 'geometry' is required (or pass a precomputed 'delaunay')")
    }
    delaunay <- geometry::delaunayn(landmarks, options = "Qt Qbb Qc Qz")
  }
  top <- matrix(t(apply(delaunay, 1, sort)), ncol = ncol(delaunay)) # m x (d_amb+1)

  # If max_dimension < ambient dimension, use the unique max_dimension-faces
  # of the Delaunay top simplices as top simplices.
  if (max_dimension < d_amb) {
    sub <- utils::combn(seq_len(d_amb + 1), max_dimension + 1, simplify = FALSE)
    top <- unique(do.call(rbind, lapply(sub, function(s) top[, s, drop = FALSE])))
  }
  d_top <- ncol(top) - 1
  m <- nrow(top)

  # grid on the standard simplex
  g <- .flood_grid(points_per_edge, d_top)
  W <- g$weights # C x (d_top+1)
  C <- nrow(W)

  # covering radii
  filt <- new.env(hash = TRUE, parent = emptyenv())
  batch_size <- max(1L, as.integer(batch_points / C))

  for (s0 in seq(1L, m, by = batch_size)) {
    s1 <- min(m, s0 + batch_size - 1L)
    nb <- s1 - s0 + 1L
    Sb <- top[s0:s1, , drop = FALSE]

    # grid points on each simplex of the batch: (nb*C) x d_amb
    P <- matrix(0, nb * C, d_amb)
    for (i in seq_len(nb)) {
      P[((i - 1L) * C + 1L):(i * C), ] <- W %*% landmarks[Sb[i, ], , drop = FALSE]
    }

    dists <- if (use_torch) .flood_nn_torch(points_t, P, device)
             else .flood_nn_cpu(points, P)
    D <- matrix(dists, nrow = nb, ncol = C, byrow = TRUE)

    # covering radius of every face of every simplex in the batch
    for (f in seq_along(g$face_rows)) {
      rows <- g$face_rows[[f]]
      verts <- g$face_vertices[[f]]
      vals <- if (length(rows) == 1L) D[, rows] else {
        M <- D[, rows, drop = FALSE]
        M[cbind(seq_len(nrow(M)), max.col(M, ties.method = "first"))]
      }
      faces_b <- Sb[, verts, drop = FALSE]                  # nb x |face|
      keys <- apply(faces_b, 1, paste, collapse = " ")
      for (i in seq_len(nb)) assign(keys[i], vals[i], envir = filt)
    }
  }

  # collect + make filtration non-decreasing
  keys <- ls(filt)
  simplices <- lapply(strsplit(keys, " ", fixed = TRUE), as.integer)
  values <- unlist(mget(keys, envir = filt), use.names = FALSE)
  dims <- lengths(simplices) - 1L

  ord <- order(dims)
  simplices <- simplices[ord]; values <- values[ord]; dims <- dims[ord]
  key_of <- vapply(simplices, paste, "", collapse = " ")
  lookup <- new.env(hash = TRUE, parent = emptyenv())
  for (i in seq_along(key_of)) assign(key_of[i], i, envir = lookup)

  for (i in seq_along(simplices)) {
    if (dims[i] == 0L) next
    s <- simplices[[i]]
    facets <- utils::combn(s, length(s) - 1L)
    for (j in seq_len(ncol(facets))) {
      fi <- get(paste(facets[, j], collapse = " "), envir = lookup)
      if (values[fi] > values[i]) values[i] <- values[fi]
    }
  }

  structure(
    list(simplices = simplices, filtration = values, landmarks = landmarks, landmark_indices = lm_idx),
    class = "flood_complex"
  )
}

#' Flood filtration in SimplicialComplex format
#'
#' Wraps \code{\link{flood_complex}} and returns the filtration in the same
#' format as \code{\link{build_filtration}}: a list of
#' \code{list(simplex = <integer vector>, t = <numeric>)}, sorted by
#' (time, dimension, lexicographic order). The result plugs directly into
#' \code{boundary_info()} / \code{extract_persistence_pairs()} /
#' \code{plot_persistence()}, as well as into the faster
#' \code{\link{flood_persistence}}.
#'
#' @param points A numeric matrix (N x d) point cloud.
#' @param landmarks Number of FPS landmarks, or an explicit landmark matrix.
#' @param ... Passed on to \code{\link{flood_complex}}.
#' @return A filtration list compatible with \code{boundary_info}.
#'
#' @export
#' @examples
#' \dontrun{
#' pts <- matrix(rnorm(2000), ncol = 2)
#' filtration <- build_flood_filtration(pts, landmarks = 30)
#' pairs <- flood_persistence(filtration)
#' }
build_flood_filtration <- function(points, landmarks, ...) {
  as_filtration(flood_complex(points, landmarks, ...))
}

#' Convert a flood_complex object to a filtration list
#'
#' @param fc A \code{"flood_complex"} object.
#' @return A filtration list compatible with \code{boundary_info}.
#'
#' @export
as_filtration <- function(fc) {
  filist <- Map(function(s, t) list(simplex = s, t = t),
                fc$simplices, fc$filtration)
  ord <- order(vapply(filist, `[[`, 0, "t"),
               lengths(lapply(filist, `[[`, "simplex")),
               vapply(filist, function(x) paste(x$simplex, collapse = "-"), ""))
  filist[ord]
}

#' Persistence pairs via sparse boundary reduction
#'
#' Computes persistence pairs from a filtration list with the standard
#' column-reduction algorithm, but on sparse columns (integer index vectors
#' over GF(2)) instead of a dense matrix. Produces the same output format as
#' \code{extract_persistence_pairs(filist, res$last_1, res$pivot_owner)} while
#' scaling to the much larger complexes produced by
#' \code{\link{build_flood_filtration}}.
#'
#' @param filist A filtration list (from \code{build_flood_filtration} or
#'   \code{build_filtration}).
#' @param max_dimension Optional maximum homology dimension to report. When
#'   set, one extra dimension is kept internally so dimension-\code{max_dimension}
#'   classes still get correct death times from their true killers, and only
#'   that extra dimension is dropped from the output - see
#'   \code{\link{restrict_filtration}}'s Details, and
#'   \code{\link{persistence_pairs}} which applies the same correction.
#'   Leave \code{NULL} (default) to report every dimension present.
#' @return A data frame with columns \code{dim}, \code{birth}, \code{death}.
#'
#' @export
flood_persistence <- function(filist, max_dimension = NULL) {
  if (!is.null(max_dimension)) {
    filist <- restrict_filtration(filist, max_dimension + 1)
  }
  n <- length(filist)
  keys <- vapply(filist, function(x) paste(x$simplex, collapse = " "), "")
  index <- new.env(hash = TRUE, parent = emptyenv())
  for (i in seq_len(n)) assign(keys[i], i, envir = index)

  cols <- vector("list", n)
  for (i in seq_len(n)) {
    s <- filist[[i]]$simplex
    if (length(s) == 1L) { cols[[i]] <- integer(0); next }
    fmat <- utils::combn(s, length(s) - 1L)
    cols[[i]] <- sort(vapply(seq_len(ncol(fmat)), function(j)
      get(paste(fmat[, j], collapse = " "), envir = index), 0L))
  }

  symdiff <- function(a, b) sort.int(c(a[!(a %in% b)], b[!(b %in% a)]))

  pivot_owner <- rep(NA_integer_, n)
  for (j in seq_len(n)) {
    col <- cols[[j]]
    repeat {
      if (length(col) == 0L) break
      piv <- col[length(col)]
      owner <- pivot_owner[piv]
      if (is.na(owner)) { pivot_owner[piv] <- j; break }
      col <- symdiff(col, cols[[owner]])
    }
    cols[[j]] <- col
  }

  dims <- lengths(lapply(filist, `[[`, "simplex")) - 1L
  ts <- vapply(filist, `[[`, 0, "t")

  killed <- which(!is.na(pivot_owner))
  pairs <- data.frame(
    dim = dims[killed],
    birth = ts[killed],
    death = ts[pivot_owner[killed]]
  )
  # zero column (positive simplex) that is never paired
  zero_cols <- which(vapply(cols, length, 0L) == 0L)
  essential <- zero_cols[is.na(pivot_owner[zero_cols])]
  if (length(essential)) {
    pairs <- rbind(pairs, data.frame(
      dim = dims[essential], birth = ts[essential], death = Inf))
  }
  rownames(pairs) <- NULL
  pairs <- pairs[order(pairs$dim, pairs$birth), ]

  if (!is.null(max_dimension)) {
    pairs <- pairs[pairs$dim <= max_dimension, , drop = FALSE]
    rownames(pairs) <- NULL
  }
  pairs
}

#' Construct a Witness Complex
#'
#' @param points A numeric matrix or data.frame of witness points \eqn{S},
#'   one point per row.
#' @param landmarks Either an integer (the number of landmarks to select from
#'   \code{points} via \code{\link{generate_landmarks}}, i.e. Farthest-Point
#'   Sampling) or an integer vector of row indices into \code{points} to use
#'   directly as landmarks \eqn{L}.
#' @param epsilon A non-negative numeric scale.
#' @param nu Landmark-distance parameter; defaults to \code{1}.
#' @return A list with:
#' \describe{
#'   \item{network}{An \code{igraph} object: the 1-skeleton on the landmark
#'     vertex set.}
#'   \item{simplices}{A list of integer vectors (indices into \code{landmarks}
#'     / \code{landmark_indices}), each the vertices of a maximal simplex.}
#'   \item{landmarks}{The landmark coordinate matrix.}
#'   \item{landmark_indices}{Row indices into \code{points}, or \code{NULL}
#'     if \code{landmarks} was given as coordinates rather than indices.}
#'   \item{edge_birth}{A symmetric matrix giving, for every pair of
#'     landmarks, the exact scale at which their edge is witnessed (used by
#'     \code{\link{build_filtration}} to time every simplex exactly, rather
#'     than only checking membership at the single scale \code{epsilon}).}
#' }
#'
#' @export
#' @examples
#' points <- matrix(rnorm(200), ncol = 2)
#' witness_complex <- WitnessComplex(points, landmarks = 15, epsilon = 0.5)
WitnessComplex <- function(points, landmarks, epsilon, nu = 1) {
  points <- as.matrix(points)

  lm_idx <- NULL
  if (length(landmarks) == 1 && is.numeric(landmarks)) {
    lm <- generate_landmarks(points, as.integer(landmarks))
    L <- lm$landmarks
    lm_idx <- lm$indices
  } else if (is.numeric(landmarks) && all(landmarks == as.integer(landmarks))) {
    lm_idx <- as.integer(landmarks)
    L <- points[lm_idx, , drop = FALSE]
  } else {
    L <- as.matrix(landmarks)
  }

  m <- nrow(L)
  D <- pairwise_dist(points, L) # N witnesses x m landmarks

  m_nu <- if (nu <= 0) {
    rep(0, nrow(D))
  } else {
    apply(D, 1, function(row) sort(row)[min(nu, length(row))])
  }

  # exact birth scale of every landmark pair: the smallest epsilon at which
  # some witness satisfies max(d(x_i,s), d(x_j,s)) <= m_nu(s) + epsilon
  edge_birth <- matrix(Inf, m, m)
  if (m >= 2) {
    for (i in 1:(m - 1)) {
      for (j in (i + 1):m) {
        b <- min(pmax(D[, i], D[, j]) - m_nu)
        edge_birth[i, j] <- edge_birth[j, i] <- b
      }
    }
  }

  edges <- which(upper.tri(edge_birth) & edge_birth <= epsilon, arr.ind = TRUE)
  complex <- build_clique_complex(m, edges)

  list(network = complex$network, simplices = complex$simplices,
       landmarks = L, landmark_indices = lm_idx, edge_birth = edge_birth)
}

#' @keywords internal
witness_scale_of_simplex <- function(edge_birth, simplex_idx) {
  if (length(simplex_idx) <= 1) return(0)
  pairs <- utils::combn(simplex_idx, 2)
  max(edge_birth[t(pairs)])
}

#' Safely compute the rank of a sparse matrix
#'
#' This helper function wraps \code{Matrix::rankMatrix()} to safely handle empty matrices (i.e., with 0 rows or columns).
#'
#' @param simplices A list of simplices representing the simplicial complex.
#' @param bound_dim The dimension of the boundary to compute the Betti number for.
#' @param tol Optional numerical tolerance to pass to \code{rankMatrix()}.
#' @return An integer representing the rank of the matrix.
#'
#' @keywords internal
#' @importFrom Matrix rankMatrix
#'
#' @export
#' @examples
#' simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))
#' betti_number(simplices, 0, tol=0.1)
betti_number <- function(
    simplices, bound_dim, tol = NULL
) {
  # Compute boundary matrices: \partial_k and \partial_{k+1}
  partial_i  <- boundary(simplices, bound_dim)
  partial_i1 <- boundary(simplices, bound_dim + 1)

  # Compute rank of \partial_{k}
  partial_i_rank <- if (bound_dim == 0) {
    0
  } else {
    safe_rank(partial_i, tol)
  }

  # Compute rank of \partial_{k+1}
  partial_i1_rank <- safe_rank(partial_i1, tol)

  # Compute Betti number: \beta_k = dim(C_k) - rank(\partial_k) - rank(\partial_{k+1})
  betti_num <- as.numeric(ncol(partial_i) - partial_i_rank - partial_i1_rank)

  return(betti_num)
}

#' @keywords internal
safe_rank <- function(mat, tol = NULL) {
  if (prod(dim(mat)) == 0) {
    return(0)
  }
  if (is.null(tol)) {
    return(rankMatrix(mat))
  } else {
    return(rankMatrix(mat, tol = tol))
  }
}

# ---------------------------------------------------------------------------
# Visualizing the matching behind a Wasserstein/bottleneck distance
#
# plot_matching() draws both persistence diagrams on one set of axes and
# connects every matched pair with a dashed line: a real-to-real match gets
# a line straight between the two points, and a point matched to the
# diagonal (i.e. effectively unmatched, the standard device described in
# R/Distance.R) gets a dashed line down to its own perpendicular projection
# onto birth = death. The underlying correspondence comes from
# diagram_matching() (R/Distance.R), which reduces to the exact same
# assignment problem wasserstein_distance()/bottleneck_distance() solve.
# ---------------------------------------------------------------------------

#' Plot the optimal matching between two persistence diagrams
#'
#' Visualizes the point correspondence that realizes the Wasserstein or
#' bottleneck distance between two persistence diagrams in a given
#' homological dimension: both diagrams' points, overlaid on the same axes,
#' joined by dashed lines to their matched partner - either a point of the
#' other diagram, or (for a point left unmatched) its own projection onto
#' the diagonal \code{birth = death}.
#'
#' @param df1,df2 Persistence diagram data frames with columns \code{dim},
#'   \code{birth}, \code{death}.
#' @param dimension Homological dimension to compare.
#' @param distance Which distance's optimal matching to visualize,
#'   \code{"wasserstein"} (default) or \code{"bottleneck"}.
#' @param p Wasserstein order, used only when \code{distance = "wasserstein"}.
#' @param ground Ground metric, \code{"L2"} or \code{"Linf"}. Defaults to
#'   each distance's own default (\code{"L2"} for Wasserstein, \code{"Linf"}
#'   for bottleneck) when left \code{NULL} - see
#'   \code{\link{wasserstein_distance}}/\code{\link{bottleneck_distance}}.
#' @param labels Legend labels identifying \code{df1} and \code{df2},
#'   e.g. \code{c("Clean", "Noisy")}.
#' @return A ggplot2 object; the title reports the resulting distance.
#'
#' @details
#' Essential (death = \code{Inf}) points are capped exactly the way
#' \code{\link{wasserstein_distance}} does (see its Details) so they take
#' part in the matching instead of being dropped, and are drawn as triangles
#' at their capped height - marked "essential" in the legend - so they stay
#' visually distinguishable from genuine finite points that happen to reach
#' that height. Matches to the diagonal are always drawn to the point's
#' perpendicular projection \code{((birth+death)/2, (birth+death)/2)}
#' regardless of \code{ground}, since that is the standard, readable way to
#' depict an unmatched point whichever ground metric produced the matching.
#'
#' @importFrom ggplot2 ggplot aes geom_point geom_segment geom_abline theme_minimal labs scale_shape_manual coord_fixed theme .data
#' @export
#' @examples
#' df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
#' df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
#' plot_matching(df1, df2, dimension = 0, distance = "wasserstein", p = 2)
plot_matching <- function(df1, df2, dimension, distance = c("wasserstein", "bottleneck"),
                           p = 2, ground = NULL, labels = c("Diagram 1", "Diagram 2")) {
  distance <- match.arg(distance)
  res <- diagram_matching(df1, df2, dimension, distance = distance, p = p, ground = ground)

  proj <- function(b, d) (b + d) / 2 # perpendicular projection onto birth = death

  seg_list <- list()

  rr <- res$matches[res$matches$type == "real-real", , drop = FALSE]
  if (nrow(rr) > 0) {
    seg_list[[length(seg_list) + 1]] <- data.frame(
      x = rr$x_birth, y = rr$x_death, xend = rr$y_birth, yend = rr$y_death
    )
  }
  xd <- res$matches[res$matches$type == "x-diagonal", , drop = FALSE]
  if (nrow(xd) > 0) {
    pr <- proj(xd$x_birth, xd$x_death)
    seg_list[[length(seg_list) + 1]] <- data.frame(x = xd$x_birth, y = xd$x_death, xend = pr, yend = pr)
  }
  yd <- res$matches[res$matches$type == "y-diagonal", , drop = FALSE]
  if (nrow(yd) > 0) {
    pr <- proj(yd$y_birth, yd$y_death)
    seg_list[[length(seg_list) + 1]] <- data.frame(x = yd$y_birth, y = yd$y_death, xend = pr, yend = pr)
  }
  seg_df <- if (length(seg_list) > 0) {
    do.call(rbind, seg_list)
  } else {
    data.frame(x = numeric(0), y = numeric(0), xend = numeric(0), yend = numeric(0))
  }

  pts_df <- rbind(
    data.frame(birth = res$X[, "birth"], death = res$X[, "death"],
               essential = res$essential_X, diagram = rep(labels[1], nrow(res$X))),
    data.frame(birth = res$Y[, "birth"], death = res$Y[, "death"],
               essential = res$essential_Y, diagram = rep(labels[2], nrow(res$Y)))
  )
  pts_df$point_type <- ifelse(pts_df$essential, "essential", "finite")
  pts_df$diagram <- factor(pts_df$diagram, levels = labels)

  vmax <- max(pts_df$birth, pts_df$death)
  pad <- max(0.08 * vmax, 1e-6)

  distance_label <- paste0(toupper(substring(distance, 1, 1)), substring(distance, 2))

  ggplot() +
    geom_segment(
      data = seg_df, aes(x = .data$x, y = .data$y, xend = .data$xend, yend = .data$yend),
      linetype = "dashed", color = "grey50", alpha = 0.8
    ) +
    geom_abline(slope = 1, intercept = 0, linetype = "dashed", color = "grey70") +
    geom_point(
      data = pts_df,
      aes(x = .data$birth, y = .data$death, color = .data$diagram, shape = .data$point_type),
      size = 3, alpha = 0.85
    ) +
    scale_shape_manual(values = c(finite = 16, essential = 17), name = NULL) +
    theme_minimal() +
    labs(
      title = sprintf("%s matching (H%d), distance = %.4g", distance_label, dimension, res$distance),
      x = "Birth", y = "Death", color = NULL
    ) +
    coord_fixed(xlim = c(min(0, -pad), vmax + pad), ylim = c(min(0, -pad), vmax + pad)) +
    theme(legend.position = "right")
}

library(Matrix)
library(gtools)
library(igraph)
library(ggplot2)
library(geometry)
library(clue)

source("./R/Faces.R")
source("./R/Boundary.R")
source("./R/Betti.R")
source("./R/ComplexUtils.R")
source("./R/VRComplex.R")
source("./R/AlphaComplex.R")
source("./R/CechComplex.R")
source("./R/FloodComplex.R") # generate_landmarks(), as_filtration()
source("./R/WitnessComplex.R")
source("./R/Filtration.R") # build_filtration()
source("./R/Persistence.R")
source("./R/Distance.R") # wasserstein_distance(), bottleneck_distance()
source("./R/CompareComplexes.R")
source("./R/ComplexDistance.R")
source("./R/PlotPD.R")
source("./R/PlotMatching.R") # diagram_matching(), plot_matching()

set.seed(11)
theta <- seq(0, 2 * pi, length.out = 30)[-30]
ring <- cbind(cos(theta), sin(theta)) + matrix(rnorm(2 * length(theta), sd = 0.03), ncol = 2)
plot(ring)

cmp <- compare_complexes(ring, methods = c("VR", "Delaunay", "Alpha", "Cech"), eps_max = 0.5)
print(cmp$summary)

cmp_w <- compare_complexes(ring, methods = "Witness", eps_max = 0.5, landmarks = 10)
print(cmp_w$summary)

# most persistent H1 bar found by each method - essential (death = Inf)
# classes are kept, not filtered, same convention as TestingComplexes.R
for (m in names(cmp$diagrams)) {
  pd <- cmp$diagrams[[m]]
  pd <- pd[pd$dim == 1, ]
  pd$pers <- pd$death - pd$birth
  best <- pd[which.max(pd$pers), ]
  cat(sprintf("%-9s longest H1 bar: birth=%.3f death=%s\n", m,
              best$birth, ifelse(is.infinite(best$death), "Inf", sprintf("%.3f", best$death))))
}


ring_noisy <- cbind(cos(theta), sin(theta)) + matrix(rnorm(2 * length(theta), sd = 0.15), ncol = 2)
plot(ring_noisy)

res_w <- complex_distance(ring, ring_noisy, method = "VR", distance = "wasserstein", dimension = 1, p = 2, eps_max = 0.6)
cat(sprintf("\nVR Wasserstein-2 (H1), clean vs. noisy ring: %.4f\n", res_w$distance))

res_b <- complex_distance(ring, ring_noisy, method = "Alpha", distance = "bottleneck", dimension = 1)
cat(sprintf("Alpha bottleneck (H1), clean vs. noisy ring: %.4f\n", res_b$distance))

# self-distance sanity check: identical point clouds are always distance 0
res_self <- complex_distance(ring, ring, method = "Delaunay", distance = "wasserstein", dimension = 1, p = 2)
cat(sprintf("Delaunay self-distance (should be 0): %.4f\n", res_self$distance))

plot_matching(res_w$diagram1, res_w$diagram2, dimension = 1,
              distance = "wasserstein", p = 2,
              labels = c("Clean ring", "Noisy ring"))

#' Construct a Cech Complex
#'
#' @param points A numeric matrix or data.frame with one point per row
#'   (columns are coordinates).
#' @param epsilon A non-negative numeric radius, used only to bound how many
#'   candidate simplices are generated (see Details) - not a promise that
#'   every returned simplex individually satisfies the Cech criterion at
#'   this scale.
#' @return A list with:
#' \describe{
#'   \item{network}{An \code{igraph} object: the 1-skeleton of the candidate
#'     graph (edges where distance \eqn{\le 2\epsilon}).}
#'   \item{simplices}{A list of integer vectors, each the vertex indices of
#'     a maximal clique of \code{network}.}
#' }
#'
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
#' cech_complex <- CechComplex(points, epsilon = 0.8)
CechComplex <- function(points, epsilon) {
  points <- as.matrix(points)
  n <- nrow(points)
  D <- pairwise_dist(points)

  edges <- which(upper.tri(D) & D <= 2 * epsilon, arr.ind = TRUE)
  build_clique_complex(n, edges)
}

#' @keywords internal
cech_scale_of_simplex <- function(points, simplex_idx) {
  if (length(simplex_idx) <= 1) return(0)
  min_enclosing_ball(points[simplex_idx, , drop = FALSE])$radius
}

#' Compare simplicial complex constructions on one point cloud
#'
#' Builds the same point cloud's filtration under several
#' \code{\link{build_filtration}} methods and summarizes, side by side, how
#' expensive each one was to build and what persistent homology it found -
#' useful for seeing directly how Cech's dimension blow-up, Alpha's
#' Delaunay-bounded dimension, VR's cheap-but-approximate cliques, and
#' Witness's landmark subsampling trade off against each other on the same
#' data (see Otter et al. (2017), Section 5.2, for the underlying
#' trade-offs).
#'
#' @param points A numeric matrix or data.frame, one point per row.
#' @param methods Character vector of methods to compare, any subset of
#'   \code{"VR"}, \code{"Delaunay"}, \code{"Alpha"}, \code{"Cech"},
#'   \code{"Witness"}. Defaults to all five.
#' @param eps_max Maximum scale, passed to \code{\link{build_filtration}}.
#'   Required whenever \code{methods} includes \code{"VR"}, \code{"Cech"} or
#'   \code{"Witness"}; optional for \code{"Alpha"} (defaults to \code{Inf});
#'   ignored for \code{"Delaunay"}.
#' @param landmarks Passed to \code{\link{build_filtration}} for
#'   \code{"Witness"}. If \code{NULL} and \code{"Witness"} is included,
#'   defaults to \code{min(30, nrow(points))} landmarks via
#'   \code{\link{generate_landmarks}}, with a message.
#' @param nu Passed to \code{\link{build_filtration}} for \code{"Witness"}.
#' @return A list with:
#' \describe{
#'   \item{summary}{A data frame, one row per method, with the number of
#'     vertices and simplices, the highest simplex dimension reached, build
#'     and persistence-computation time in seconds, and the number of
#'     finite/essential persistence pairs found.}
#'   \item{diagrams}{A named list of persistence diagram data frames (one per
#'     method, from \code{\link{persistence_pairs}}), for further comparison
#'     (e.g. with \code{\link{wasserstein_distance}}/
#'     \code{\link{bottleneck_distance}}, or \code{\link{plot_persistence}}).}
#'   \item{filtrations}{A named list of the raw filtration lists.}
#' }
#'
#' @export
#' @examples
#' points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1, 0.5, 0.5), ncol = 2, byrow = TRUE)
#' cmp <- compare_complexes(points, methods = c("VR", "Alpha", "Delaunay"), eps_max = 1.2)
#' cmp$summary
compare_complexes <- function(points,
                               methods = c("VR", "Delaunay", "Alpha", "Cech", "Witness"),
                               eps_max = NULL,
                               landmarks = NULL,
                               nu = 1) {
  points <- as.matrix(points)
  methods <- match.arg(methods,
                        choices = c("VR", "Delaunay", "Alpha", "Cech", "Witness"),
                        several.ok = TRUE)

  if (any(methods %in% c("VR", "Cech", "Witness")) && is.null(eps_max)) {
    stop("eps_max is required for methods 'VR', 'Cech' and 'Witness'.")
  }
  if ("Witness" %in% methods && is.null(landmarks)) {
    landmarks <- min(30L, nrow(points))
    message(sprintf(
      "No landmarks supplied for method = 'Witness'; using %d landmarks via generate_landmarks().",
      landmarks
    ))
  }

  filtrations <- list()
  diagrams <- list()
  rows <- vector("list", length(methods))

  for (i in seq_along(methods)) {
    m <- methods[i]
    build_time <- system.time(
      filt <- if (m == "Delaunay") {
        build_filtration(points, method = m)
      } else if (m == "Witness") {
        build_filtration(points, method = m, eps_max = eps_max, landmarks = landmarks, nu = nu)
      } else {
        build_filtration(points, method = m, eps_max = eps_max)
      }
    )[["elapsed"]]

    pers_time <- system.time(diag <- persistence_pairs(filt))[["elapsed"]]

    dims <- lengths(lapply(filt, `[[`, "simplex")) - 1L

    filtrations[[m]] <- filt
    diagrams[[m]] <- diag
    rows[[i]] <- data.frame(
      method = m,
      n_vertices = sum(dims == 0),
      n_simplices = length(filt),
      max_dimension = max(dims),
      build_time_sec = round(build_time, 4),
      persistence_time_sec = round(pers_time, 4),
      n_finite_pairs = sum(is.finite(diag$death)),
      n_essential_pairs = sum(is.infinite(diag$death)),
      stringsAsFactors = FALSE
    )
  }

  summary <- do.call(rbind, rows)
  rownames(summary) <- NULL

  list(summary = summary, diagrams = diagrams, filtrations = filtrations)
}

#' @keywords internal
diagram_points <- function(df, dimension) {
  sub <- df[df$dim == dimension, , drop = FALSE]
  cbind(birth = sub$birth, death = sub$death)
}

#' @keywords internal
cap_essential <- function(X, Y, dimension) {
  finite_vals <- c(X[is.finite(X)], Y[is.finite(Y)])
  cap <- if (length(finite_vals) > 0) 2 * max(finite_vals) else 1
  if (!is.finite(cap) || cap <= 0) cap <- 1

  n_essential <- sum(!is.finite(X)) + sum(!is.finite(Y))
  if (n_essential > 0) {
    message(sprintf(
      "Capping %d essential (death = Inf) pair(s) in dimension %d at death = %.4g so they are compared by birth time rather than discarded.",
      n_essential, dimension, cap
    ))
  }

  X[!is.finite(X)] <- cap
  Y[!is.finite(Y)] <- cap
  list(X = X, Y = Y)
}

#' @keywords internal
ground_dist <- function(X, Y, ground = c("L2", "Linf")) {
  ground <- match.arg(ground)
  if (ground == "L2") return(pairwise_dist(X, Y))
  # Linf (Chebyshev): max coordinate-wise difference
  n <- nrow(X); m <- nrow(Y)
  out <- matrix(0, n, m)
  for (k in seq_len(ncol(X))) out <- pmax(out, abs(outer(X[, k], Y[, k], "-")))
  out
}

#' @keywords internal
dist_to_diagonal <- function(X, ground = c("L2", "Linf")) {
  ground <- match.arg(ground)
  gap <- abs(X[, "death"] - X[, "birth"])
  if (ground == "L2") gap / sqrt(2) else gap / 2
}

#' Build the augmented (n+m) x (n+m) assignment cost matrix for two diagrams
#'
#' @param X,Y Matrices with columns \code{birth}, \code{death} (as produced
#'   by the internal \code{diagram_points()} helper).
#' @param ground Ground metric on the birth-death plane, \code{"L2"} or
#'   \code{"Linf"}.
#' @param power Exponent applied to every ground distance before it enters
#'   the matrix (\code{p} for a \eqn{p}-Wasserstein distance; use \code{1}
#'   for bottleneck, which works with raw distances and takes a max instead
#'   of a sum).
#' @return A square numeric matrix of size \code{nrow(X) + nrow(Y)}.
#'
#' @keywords internal
augmented_cost_matrix <- function(X, Y, ground = c("L2", "Linf"), power = 1) {
  ground <- match.arg(ground)
  n <- nrow(X); m <- nrow(Y)

  real_real <- if (n > 0 && m > 0) ground_dist(X, Y, ground) else matrix(0, n, m)
  dX <- if (n > 0) dist_to_diagonal(X, ground) else numeric(0)
  dY <- if (m > 0) dist_to_diagonal(Y, ground) else numeric(0)

  real_real <- real_real^power
  dX <- dX^power
  dY <- dY^power

  top <- cbind(real_real, matrix(dX, nrow = n, ncol = n))
  bottom <- cbind(matrix(dY, nrow = m, ncol = m, byrow = TRUE), matrix(0, nrow = m, ncol = n))
  rbind(top, bottom)
}

#' @keywords internal
wasserstein_assignment <- function(C, p) {
  assignment <- clue::solve_LSAP(C)
  total <- sum(C[cbind(seq_len(nrow(C)), assignment)])
  list(distance = total^(1 / p), assignment = assignment)
}

#' Wasserstein distance between two persistence diagrams
#'
#' The \eqn{p}-Wasserstein distance between the points of two persistence
#' diagrams in a given homological dimension, allowing points to be matched
#' to the diagonal (Cohen-Steiner, Edelsbrunner, Harer and Mileyko (2010),
#' "Lipschitz Functions Have \eqn{L_p}-Stable Persistence"). Essential
#' (death = \code{Inf}) classes are kept rather than dropped - see Details.
#' Computed exactly via the Hungarian algorithm (\code{clue::solve_LSAP}) on
#' the augmented assignment problem described in
#' \code{\link{augmented_cost_matrix}}.
#'
#' @param df1,df2 Persistence diagram data frames with columns \code{dim},
#'   \code{birth}, \code{death} (e.g. the output of
#'   \code{extract_persistence_pairs()} or \code{persistence_pairs()}).
#' @param dimension Homological dimension to compare.
#' @param p Wasserstein order (\eqn{p \ge 1}). Defaults to \code{2}.
#' @param ground Ground metric on the birth-death plane used to measure the
#'   distance between two points (or a point and the diagonal): \code{"L2"}
#'   (Euclidean, default) or \code{"Linf"} (Chebyshev).
#' @return A single non-negative number, the \eqn{p}-Wasserstein distance.
#'
#' @importFrom clue solve_LSAP
#' @export
#' @examples
#' df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
#' df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
#' wasserstein_distance(df1, df2, dimension = 0, p = 2)
wasserstein_distance <- function(df1, df2, dimension = 0, p = 2, ground = c("L2", "Linf")) {
  ground <- match.arg(ground)
  if (p < 1) stop("p must be >= 1.")

  X <- diagram_points(df1, dimension)
  Y <- diagram_points(df2, dimension)
  if (nrow(X) == 0 && nrow(Y) == 0) return(0)
  capped <- cap_essential(X, Y, dimension)
  X <- capped$X; Y <- capped$Y

  C <- augmented_cost_matrix(X, Y, ground = ground, power = p)
  wasserstein_assignment(C, p)$distance
}

#' @keywords internal
bipartite_perfect_matching_exists <- function(adj) {
  n <- nrow(adj)
  match_right <- integer(n) # 0 = unmatched

  augment <- function(u, visited) {
    for (v in which(adj[u, ])) {
      if (visited[v]) next
      visited[v] <- TRUE
      if (match_right[v] == 0L || augment(match_right[v], visited)) {
        match_right[v] <<- u
        return(TRUE)
      }
    }
    FALSE
  }

  for (u in seq_len(n)) {
    if (!augment(u, logical(n))) return(FALSE)
  }
  TRUE
}

#' @keywords internal
bipartite_maximum_matching <- function(adj) {
  # Same augmenting-path search as bipartite_perfect_matching_exists(), but
  # runs every row to completion instead of bailing out on the first row
  # that fails to match - used to recover the actual matching (not just
  # existence) once a feasible threshold has already been found.
  N <- nrow(adj)
  match_right <- integer(N) # 0 = unmatched; match_right[v] = matched row u

  augment <- function(u, visited) {
    for (v in which(adj[u, ])) {
      if (visited[v]) next
      visited[v] <- TRUE
      if (match_right[v] == 0L || augment(match_right[v], visited)) {
        match_right[v] <<- u
        return(TRUE)
      }
    }
    FALSE
  }

  for (u in seq_len(N)) augment(u, logical(N))
  match_right
}

#' @keywords internal
bottleneck_threshold <- function(C) {
  candidates <- sort(unique(c(0, as.vector(C))))
  lo <- 1L
  hi <- length(candidates)
  while (lo < hi) {
    mid <- (lo + hi) %/% 2L
    adj <- C <= candidates[mid] + 1e-9
    if (bipartite_perfect_matching_exists(adj)) hi <- mid else lo <- mid + 1L
  }
  candidates[lo]
}

#' Bottleneck distance between two persistence diagrams
#'
#' The bottleneck distance between the points of two persistence diagrams in
#' a given homological dimension, allowing points to be matched to the
#' diagonal: the infimum, over all matchings, of the largest single
#' point-to-point distance (Cohen-Steiner, Edelsbrunner and Harer (2007),
#' "Stability of Persistence Diagrams") - the \eqn{p \to \infty} limit of
#' \code{\link{wasserstein_distance}}. Essential (death = \code{Inf})
#' classes are kept rather than dropped - see
#' \code{\link{wasserstein_distance}}'s Details for why and how. Computed
#' exactly by binary search over the (finitely many) candidate distance
#' values for the smallest one admitting a perfect matching, checked with a
#' standard augmenting-path bipartite matcher, on the same augmented
#' assignment problem \code{\link{wasserstein_distance}} uses (see
#' \code{\link{augmented_cost_matrix}}).
#'
#' @param df1,df2 Persistence diagram data frames with columns \code{dim},
#'   \code{birth}, \code{death}.
#' @param dimension Homological dimension to compare.
#' @param ground Ground metric on the birth-death plane: \code{"L2"} or
#'   \code{"Linf"} (Chebyshev, the convention used by the reference above;
#'   default).
#' @return A single non-negative number, the bottleneck distance.
#'
#' @export
#' @examples
#' df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
#' df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
#' bottleneck_distance(df1, df2, dimension = 0)
bottleneck_distance <- function(df1, df2, dimension = 0, ground = c("Linf", "L2")) {
  ground <- match.arg(ground)

  X <- diagram_points(df1, dimension)
  Y <- diagram_points(df2, dimension)
  if (nrow(X) == 0 && nrow(Y) == 0) return(0)
  capped <- cap_essential(X, Y, dimension)
  X <- capped$X; Y <- capped$Y

  C <- augmented_cost_matrix(X, Y, ground = ground, power = 1)
  bottleneck_threshold(C)
}

#' Optimal point matching between two persistence diagrams
#'
#' Computes the same optimal matching that
#' \code{\link{wasserstein_distance}}/\code{\link{bottleneck_distance}}
#' reduce to internally, and returns it decoded into a readable table
#' instead of just the distance - what \code{\link{plot_matching}} draws.
#' Every point of both diagrams appears in exactly one row: matched to a
#' point of the other diagram (\code{type = "real-real"}), or matched to the
#' diagonal, i.e. effectively unmatched (\code{type = "x-diagonal"} for a
#' point of \code{df1}, \code{"y-diagonal"} for a point of \code{df2}).
#'
#' @param df1,df2 Persistence diagram data frames with columns \code{dim},
#'   \code{birth}, \code{death}.
#' @param dimension Homological dimension to compare.
#' @param distance Which distance's optimal matching to compute,
#'   \code{"wasserstein"} (default) or \code{"bottleneck"}.
#' @param p Wasserstein order, used only when \code{distance = "wasserstein"}.
#' @param ground Ground metric on the birth-death plane, \code{"L2"} or
#'   \code{"Linf"}. Defaults to each distance's own default (\code{"L2"} for
#'   Wasserstein, \code{"Linf"} for bottleneck) when left \code{NULL}.
#' @return A list with \code{distance} (the matching's cost, equal to what
#'   \code{\link{wasserstein_distance}}/\code{\link{bottleneck_distance}}
#'   would return), \code{matches} (a data frame with columns
#'   \code{x_birth, x_death, x_essential, y_birth, y_death, y_essential,
#'   type}; a \code{NA} pair on one side means that row's point matched the
#'   diagonal), and \code{X}, \code{Y}, \code{essential_X}, \code{essential_Y}
#'   (the two diagrams' points after the same essential-pair capping
#'   \code{\link{wasserstein_distance}} uses - see its Details - and which of
#'   them were essential before capping).
#'
#' @export
#' @examples
#' df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
#' df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
#' diagram_matching(df1, df2, dimension = 0, distance = "wasserstein", p = 2)
diagram_matching <- function(df1, df2, dimension, distance = c("wasserstein", "bottleneck"),
                              p = 2, ground = NULL) {
  distance <- match.arg(distance)
  if (is.null(ground)) {
    ground <- if (distance == "bottleneck") "Linf" else "L2"
  } else {
    ground <- match.arg(ground, c("L2", "Linf"))
  }
  if (distance == "wasserstein" && p < 1) stop("p must be >= 1.")

  X <- diagram_points(df1, dimension)
  Y <- diagram_points(df2, dimension)
  n <- nrow(X); m <- nrow(Y)
  if (n == 0 && m == 0) {
    stop(sprintf("No points in dimension %d in either diagram - nothing to match.", dimension))
  }

  essential_X <- if (n > 0) !is.finite(X[, "death"]) else logical(0)
  essential_Y <- if (m > 0) !is.finite(Y[, "death"]) else logical(0)

  capped <- cap_essential(X, Y, dimension)
  X <- capped$X; Y <- capped$Y

  power <- if (distance == "wasserstein") p else 1
  C <- augmented_cost_matrix(X, Y, ground = ground, power = power)

  if (distance == "wasserstein") {
    solved <- wasserstein_assignment(C, p)
    dist_val <- solved$distance
    row_to_col <- solved$assignment # row_to_col[i] = column matched to row i
  } else {
    dist_val <- bottleneck_threshold(C)
    adj <- C <= dist_val + 1e-9
    match_right <- bipartite_maximum_matching(adj) # match_right[v] = row matched to column v
    row_to_col <- integer(length(match_right))
    row_to_col[match_right] <- seq_along(match_right)
  }

  # Decode: rows 1..n are df1's real points, rows n+1..n+m are diagonal
  # slots that let df2's real points match the diagonal; columns 1..m are
  # df2's real points, columns m+1..m+n are diagonal slots for df1's real
  # points (see augmented_cost_matrix()).
  matches <- vector("list", n + m)
  k <- 0L
  for (i in seq_len(n)) {
    j <- row_to_col[i]
    k <- k + 1L
    if (j <= m) {
      matches[[k]] <- data.frame(
        x_birth = X[i, "birth"], x_death = X[i, "death"], x_essential = essential_X[i],
        y_birth = Y[j, "birth"], y_death = Y[j, "death"], y_essential = essential_Y[j],
        type = "real-real"
      )
    } else {
      matches[[k]] <- data.frame(
        x_birth = X[i, "birth"], x_death = X[i, "death"], x_essential = essential_X[i],
        y_birth = NA_real_, y_death = NA_real_, y_essential = NA,
        type = "x-diagonal"
      )
    }
  }
  if (m > 0) {
    for (v in seq_len(m)) {
      row_idx <- n + v
      j <- row_to_col[row_idx]
      if (j <= m) {
        # a diagonal-slot row matched to column j means df2's point j went
        # to the diagonal; the row's own identity (v) is not meaningful
        k <- k + 1L
        matches[[k]] <- data.frame(
          x_birth = NA_real_, x_death = NA_real_, x_essential = NA,
          y_birth = Y[j, "birth"], y_death = Y[j, "death"], y_essential = essential_Y[j],
          type = "y-diagonal"
        )
      }
      # else: this diagonal slot matched a df1 diagonal slot (cost 0,
      # neither side a real point) - nothing to report
    }
  }

  list(
    distance = dist_val,
    matches = do.call(rbind, matches[seq_len(k)]),
    X = X, Y = Y, essential_X = essential_X, essential_Y = essential_Y
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VRComplex.R
\name{VietorisRipsComplex}
\alias{VietorisRipsComplex}
\title{Construct a Vietoris–Rips Complex (1-skeleton + maximal simplices)}
\usage{
VietorisRipsComplex(points, epsilon)
}
\arguments{
\item{points}{A numeric matrix or data.frame with one point per row (columns are coordinates).}

\item{epsilon}{A positive numeric threshold; connect points with distance < \eqn{\epsilon}.}
}
\value{
A list with:
\describe{
  \item{network}{An \code{igraph} object representing the 1-skeleton.}
  \item{simplices}{A list of integer vectors, each the vertex indices of a maximal simplex.}
}
}
\description{
Construct a Vietoris–Rips Complex (1-skeleton + maximal simplices)
}
\details{
The Vietoris–Rips complex at scale \eqn{\epsilon} includes a simplex for every
finite set of points with pairwise distances < \eqn{\epsilon}. This function
constructs the 1-skeleton (edges only) and then uses maximal cliques in that
graph as the maximal simplices.
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
epsilon <- 1.5
vr_complex <- VietorisRipsComplex(points, epsilon)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Persistence.R
\name{boundary_info}
\alias{boundary_info}
\title{Get the boundary matrix and its reduction information in matrix form}
\usage{
boundary_info(filist, max_dimension = NULL)
}
\arguments{
\item{filist}{Filtration list, each element includes simplex and time.}

\item{max_dimension}{Optional maximum homology dimension to eventually
report via \code{\link{extract_persistence_pairs}}. When set,
\code{filist} is first restricted to \code{max_dimension + 1} (one
extra dimension, kept only so dimension-\code{max_dimension} classes
are correctly killed - see \code{\link{restrict_filtration}}'s
Details) before the boundary matrix is built and reduced. Leave
\code{NULL} (default) to fall back to \code{filist}'s own
\code{"max_dimension"} attribute (set automatically when it came from
\code{\link{build_filtration}} with \code{max_dimension} supplied
there), or to use \code{filist} uncapped if that attribute is also
absent.}
}
\value{
A list containing the boundary matrix, the last boundary row, the
  pivot owner for persistence extraction, and \code{filist} - the exact
  (possibly \code{max_dimension}-restricted) filtration list the other
  three elements were computed from, re-tagged with the same
  \code{"max_dimension"} attribute so \code{\link{extract_persistence_pairs}}
  can auto-detect it too. Always pass THIS \code{filist} back into
  \code{\link{extract_persistence_pairs}}, not your original one - see its
  Details for why.
}
\description{
Get the boundary matrix and its reduction information in matrix form
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
res <- boundary_info(filtration)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotPD.R
\name{plot_landscape}
\alias{plot_landscape}
\title{Plot a Persistence Landscape}
\usage{
plot_landscape(landscape_df)
}
\arguments{
\item{landscape_df}{Data frame from \code{persistence_landscape()}, with
columns \code{t}, \code{k}, \code{value}.}
}
\value{
A ggplot2 object with one line per landscape level
  \eqn{\lambda(k, \cdot)}.
}
\description{
Plot a Persistence Landscape
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
res <- boundary_info(filtration)
pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
landscape <- persistence_landscape(pairs, dimension = 0)
plot_landscape(landscape)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CompareComplexes.R
\name{compare_complexes}
\alias{compare_complexes}
\title{Compare simplicial complex constructions on one point cloud}
\usage{
compare_complexes(
  points,
  methods = c("VR", "Delaunay", "Alpha", "Cech", "Witness"),
  eps_max = NULL,
  landmarks = NULL,
  nu = 1
)
}
\arguments{
\item{points}{A numeric matrix or data.frame, one point per row.}

\item{methods}{Character vector of methods to compare, any subset of
\code{"VR"}, \code{"Delaunay"}, \code{"Alpha"}, \code{"Cech"},
\code{"Witness"}. Defaults to all five.}

\item{eps_max}{Maximum scale, passed to \code{\link{build_filtration}}.
Required whenever \code{methods} includes \code{"VR"}, \code{"Cech"} or
\code{"Witness"}; optional for \code{"Alpha"} (defaults to \code{Inf});
ignored for \code{"Delaunay"}.}

\item{landmarks}{Passed to \code{\link{build_filtration}} for
\code{"Witness"}. If \code{NULL} and \code{"Witness"} is included,
defaults to \code{min(30, nrow(points))} landmarks via
\code{\link{generate_landmarks}}, with a message.}

\item{nu}{Passed to \code{\link{build_filtration}} for \code{"Witness"}.}
}
\value{
A list with:
\describe{
  \item{summary}{A data frame, one row per method, with the number of
    vertices and simplices, the highest simplex dimension reached, build
    and persistence-computation time in seconds, and the number of
    finite/essential persistence pairs found.}
  \item{diagrams}{A named list of persistence diagram data frames (one per
    method, from \code{\link{persistence_pairs}}), for further comparison
    (e.g. with \code{\link{wasserstein_distance}}/
    \code{\link{bottleneck_distance}}, or \code{\link{plot_persistence}}).}
  \item{filtrations}{A named list of the raw filtration lists.}
}
}
\description{
Builds the same point cloud's filtration under several
\code{\link{build_filtration}} methods and summarizes, side by side, how
expensive each one was to build and what persistent homology it found -
useful for seeing directly how Cech's dimension blow-up, Alpha's
Delaunay-bounded dimension, VR's cheap-but-approximate cliques, and
Witness's landmark subsampling trade off against each other on the same
data (see Otter et al. (2017), Section 5.2, for the underlying
trade-offs).
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1, 0.5, 0.5), ncol = 2, byrow = TRUE)
cmp <- compare_complexes(points, methods = c("VR", "Alpha", "Delaunay"), eps_max = 1.2)
cmp$summary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FloodComplex.R
\name{flood_complex}
\alias{flood_complex}
\title{Construct a Flood complex}
\usage{
flood_complex(
  points,
  landmarks,
  max_dimension = NULL,
  points_per_edge = 30,
  backend = c("auto", "cpu", "torch"),
  batch_points = 2^22,
  delaunay = NULL
)
}
\arguments{
\item{points}{A numeric matrix (N x d) of witness points.}

\item{landmarks}{Either an integer (number of FPS landmarks) or a numeric matrix (N_l x d) of explicit landmark coordinates.}

\item{max_dimension}{Top dimension of the simplices. Defaults to d.}

\item{points_per_edge}{Grid resolution per simplex edge (accuracy vs. speed trade-off). Defaults to 30, as in flooder.}

\item{backend}{One of \code{"auto"}, \code{"cpu"}, \code{"torch"}.
\code{"cpu"} uses a kd-tree (RANN). \code{"torch"} uses the torch R package and runs on CUDA when available. \code{"auto"} picks torch only if a CUDA device is present, else the kd-tree.}

\item{batch_points}{Maximum number of grid points processed per batch (bounds memory). Defaults to 2^22.}

\item{delaunay}{Optional precomputed Delaunay triangulation of the landmarks: an m x (d+1) integer matrix of 1-based landmark indices. If NULL (default), computed via \code{geometry::delaunayn}.}
}
\value{
A list of class \code{"flood_complex"} with elements
  \code{simplices} (list of integer vectors, landmark indices),
  \code{filtration} (numeric vector of flood times),
  \code{landmarks} (matrix), \code{landmark_indices} (or NULL).
}
\description{
Builds the Flood complex of a point cloud: a Delaunay complex on a set of
landmarks whose simplices are filtered by their covering radius with
respect to the full point cloud ("flood time").
}
\examples{
\dontrun{
pts <- matrix(rnorm(3000), ncol = 2)
fc <- flood_complex(pts, landmarks = 40)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Distance.R
\name{augmented_cost_matrix}
\alias{augmented_cost_matrix}
\title{Build the augmented (n+m) x (n+m) assignment cost matrix for two diagrams}
\usage{
augmented_cost_matrix(X, Y, ground = c("L2", "Linf"), power = 1)
}
\arguments{
\item{X, Y}{Matrices with columns \code{birth}, \code{death} (as produced
by the internal \code{diagram_points()} helper).}

\item{ground}{Ground metric on the birth-death plane, \code{"L2"} or
\code{"Linf"}.}

\item{power}{Exponent applied to every ground distance before it enters
the matrix (\code{p} for a \eqn{p}-Wasserstein distance; use \code{1}
for bottleneck, which works with raw distances and takes a max instead
of a sum).}
}
\value{
A square numeric matrix of size \code{nrow(X) + nrow(Y)}.
}
\description{
Build the augmented (n+m) x (n+m) assignment cost matrix for two diagrams
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotMatching.R
\name{plot_matching}
\alias{plot_matching}
\title{Plot the optimal matching between two persistence diagrams}
\usage{
plot_matching(
  df1,
  df2,
  dimension,
  distance = c("wasserstein", "bottleneck"),
  p = 2,
  ground = NULL,
  labels = c("Diagram 1", "Diagram 2")
)
}
\arguments{
\item{df1, df2}{Persistence diagram data frames with columns \code{dim},
\code{birth}, \code{death}.}

\item{dimension}{Homological dimension to compare.}

\item{distance}{Which distance's optimal matching to visualize,
\code{"wasserstein"} (default) or \code{"bottleneck"}.}

\item{p}{Wasserstein order, used only when \code{distance = "wasserstein"}.}

\item{ground}{Ground metric, \code{"L2"} or \code{"Linf"}. Defaults to
each distance's own default (\code{"L2"} for Wasserstein, \code{"Linf"}
for bottleneck) when left \code{NULL} - see
\code{\link{wasserstein_distance}}/\code{\link{bottleneck_distance}}.}

\item{labels}{Legend labels identifying \code{df1} and \code{df2},
e.g. \code{c("Clean", "Noisy")}.}
}
\value{
A ggplot2 object; the title reports the resulting distance.
}
\description{
Visualizes the point correspondence that realizes the Wasserstein or
bottleneck distance between two persistence diagrams in a given
homological dimension: both diagrams' points, overlaid on the same axes,
joined by dashed lines to their matched partner - either a point of the
other diagram, or (for a point left unmatched) its own projection onto
the diagonal \code{birth = death}.
}
\details{
Essential (death = \code{Inf}) points are capped exactly the way
\code{\link{wasserstein_distance}} does (see its Details) so they take
part in the matching instead of being dropped, and are drawn as triangles
at their capped height - marked "essential" in the legend - so they stay
visually distinguishable from genuine finite points that happen to reach
that height. Matches to the diagonal are always drawn to the point's
perpendicular projection \code{((birth+death)/2, (birth+death)/2)}
regardless of \code{ground}, since that is the standard, readable way to
depict an unmatched point whichever ground metric produced the matching.
}
\examples{
df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
plot_matching(df1, df2, dimension = 0, distance = "wasserstein", p = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Persistence.R
\name{extract_persistence_pairs}
\alias{extract_persistence_pairs}
\title{This function extracts the persistence from combining the boundary matrix and its filtration}
\usage{
extract_persistence_pairs(filist, last_1, pivot_owner, max_dimension = NULL)
}
\arguments{
\item{filist}{Filtration list, each element includes simplex and time.
When \code{boundary_info()} was called with \code{max_dimension} set,
this MUST be \code{res$filist} (the restricted list it returned), not
your original filtration - \code{last_1}/\code{pivot_owner} are indexed
positionally against whatever \code{filist} \code{boundary_info()}
actually used, so passing a different-length \code{filist} here would
silently pair the wrong simplices. See Details.}

\item{last_1}{The last 1 row index for each column in boundary matrix (after reduction).}

\item{pivot_owner}{The column index owning the pivot row.}

\item{max_dimension}{Optional maximum homology dimension to report. Set
this to the SAME value passed to \code{boundary_info()} (which kept one
extra dimension internally for correct killers - see
\code{\link{restrict_filtration}}'s Details); only that extra dimension
is dropped here. Leave \code{NULL} (default) to fall back to
\code{filist}'s own \code{"max_dimension"} attribute (present when
\code{filist} is \code{res$filist} from a \code{boundary_info()} call
that used \code{max_dimension}, directly or via its own fallback to
\code{\link{build_filtration}}'s attribute), or to report every
dimension present in \code{filist} if that attribute is also absent.}
}
\value{
A data frame with columns: dimension, birth, and death.
}
\description{
This function extracts the persistence from combining the boundary matrix and its filtration
}
\details{
\code{boundary_info()} and \code{extract_persistence_pairs()} are two
halves of one computation - \code{last_1}/\code{pivot_owner} only mean
anything relative to the exact \code{filist} \code{boundary_info()} used
internally. Whenever \code{max_dimension} is involved, always call as:
\preformatted{
res <- boundary_info(filtration, max_dimension = k)
pairs <- extract_persistence_pairs(res$filist, res$last_1, res$pivot_owner,
                                    max_dimension = k)
}
A length mismatch between \code{filist} and \code{last_1}/\code{pivot_owner}
is refused with an error rather than silently producing wrong pairs - see
\code{\link{persistence_pairs}} for a one-call alternative that cannot run
into this.
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
res <- boundary_info(filtration)
pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Distance.R
\name{bottleneck_distance}
\alias{bottleneck_distance}
\title{Bottleneck distance between two persistence diagrams}
\usage{
bottleneck_distance(df1, df2, dimension = 0, ground = c("Linf", "L2"))
}
\arguments{
\item{df1, df2}{Persistence diagram data frames with columns \code{dim},
\code{birth}, \code{death}.}

\item{dimension}{Homological dimension to compare.}

\item{ground}{Ground metric on the birth-death plane: \code{"L2"} or
\code{"Linf"} (Chebyshev, the convention used by the reference above;
default).}
}
\value{
A single non-negative number, the bottleneck distance.
}
\description{
The bottleneck distance between the points of two persistence diagrams in
a given homological dimension, allowing points to be matched to the
diagonal: the infimum, over all matchings, of the largest single
point-to-point distance (Cohen-Steiner, Edelsbrunner and Harer (2007),
"Stability of Persistence Diagrams") - the \eqn{p \to \infty} limit of
\code{\link{wasserstein_distance}}. Essential (death = \code{Inf})
classes are kept rather than dropped - see
\code{\link{wasserstein_distance}}'s Details for why and how. Computed
exactly by binary search over the (finitely many) candidate distance
values for the smallest one admitting a perfect matching, checked with a
standard augmenting-path bipartite matcher, on the same augmented
assignment problem \code{\link{wasserstein_distance}} uses (see
\code{\link{augmented_cost_matrix}}).
}
\examples{
df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
bottleneck_distance(df1, df2, dimension = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AlphaComplex.R
\name{DelaunayComplex}
\alias{DelaunayComplex}
\title{Construct a Delaunay Complex}
\usage{
DelaunayComplex(points)
}
\arguments{
\item{points}{A numeric matrix or data.frame with one point per row
(columns are coordinates); must be in "general position".}
}
\value{
A list of class \code{c("delaunay_complex", "alpha_complex")};
  see \code{\link{AlphaComplex}} for the element description.
}
\description{
Construct a Delaunay Complex
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
delaunay_complex <- DelaunayComplex(points)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AlphaComplex.R
\name{AlphaComplex}
\alias{AlphaComplex}
\title{Construct an Alpha Complex}
\usage{
AlphaComplex(points, epsilon = Inf)
}
\arguments{
\item{points}{A numeric matrix or data.frame with one point per row
(columns are coordinates); must be in "general position" (see
\code{\link{circumsphere}}).}

\item{epsilon}{The alpha-complex scale. A simplex is included once its
alpha value is at most \code{epsilon}. Defaults to \code{Inf}, which
gives the full Delaunay complex (see \code{\link{DelaunayComplex}}).}
}
\value{
A list of class \code{"alpha_complex"} with elements:
\describe{
  \item{simplices}{A list of integer vectors, every included simplex (all
    dimensions, not just the maximal ones - membership in the alpha
    complex is not simply generated by cliques).}
  \item{filtration}{A numeric vector, the alpha value of each simplex in
    \code{simplices}, in the same order.}
}
Pass the result to \code{\link{as_filtration}} for the same
\code{list(simplex=, t=)} format used by \code{\link{build_filtration}}
and every persistence function, or just call
\code{build_filtration(points, method = "Alpha", eps_max = epsilon)}.
}
\description{
Construct an Alpha Complex
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
alpha_complex <- AlphaComplex(points, epsilon = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CechComplex.R
\name{CechComplex}
\alias{CechComplex}
\title{Construct a Cech Complex}
\usage{
CechComplex(points, epsilon)
}
\arguments{
\item{points}{A numeric matrix or data.frame with one point per row
(columns are coordinates).}

\item{epsilon}{A non-negative numeric radius, used only to bound how many
candidate simplices are generated (see Details) - not a promise that
every returned simplex individually satisfies the Cech criterion at
this scale.}
}
\value{
A list with:
\describe{
  \item{network}{An \code{igraph} object: the 1-skeleton of the candidate
    graph (edges where distance \eqn{\le 2\epsilon}).}
  \item{simplices}{A list of integer vectors, each the vertex indices of
    a maximal clique of \code{network}.}
}
}
\description{
Construct a Cech Complex
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
cech_complex <- CechComplex(points, epsilon = 0.8)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FloodComplex.R
\name{build_flood_filtration}
\alias{build_flood_filtration}
\title{Flood filtration in SimplicialComplex format}
\usage{
build_flood_filtration(points, landmarks, ...)
}
\arguments{
\item{points}{A numeric matrix (N x d) point cloud.}

\item{landmarks}{Number of FPS landmarks, or an explicit landmark matrix.}

\item{...}{Passed on to \code{\link{flood_complex}}.}
}
\value{
A filtration list compatible with \code{boundary_info}.
}
\description{
Wraps \code{\link{flood_complex}} and returns the filtration in the same
format as \code{\link{build_filtration}}: a list of
\code{list(simplex = <integer vector>, t = <numeric>)}, sorted by
(time, dimension, lexicographic order). The result plugs directly into
\code{boundary_info()} / \code{extract_persistence_pairs()} /
\code{plot_persistence()}, as well as into the faster
\code{\link{flood_persistence}}.
}
\examples{
\dontrun{
pts <- matrix(rnorm(2000), ncol = 2)
filtration <- build_flood_filtration(pts, landmarks = 30)
pairs <- flood_persistence(filtration)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WitnessComplex.R
\name{WitnessComplex}
\alias{WitnessComplex}
\title{Construct a Witness Complex}
\usage{
WitnessComplex(points, landmarks, epsilon, nu = 1)
}
\arguments{
\item{points}{A numeric matrix or data.frame of witness points \eqn{S},
one point per row.}

\item{landmarks}{Either an integer (the number of landmarks to select from
\code{points} via \code{\link{generate_landmarks}}, i.e. Farthest-Point
Sampling) or an integer vector of row indices into \code{points} to use
directly as landmarks \eqn{L}.}

\item{epsilon}{A non-negative numeric scale.}

\item{nu}{Landmark-distance parameter; defaults to \code{1}.}
}
\value{
A list with:
\describe{
  \item{network}{An \code{igraph} object: the 1-skeleton on the landmark
    vertex set.}
  \item{simplices}{A list of integer vectors (indices into \code{landmarks}
    / \code{landmark_indices}), each the vertices of a maximal simplex.}
  \item{landmarks}{The landmark coordinate matrix.}
  \item{landmark_indices}{Row indices into \code{points}, or \code{NULL}
    if \code{landmarks} was given as coordinates rather than indices.}
  \item{edge_birth}{A symmetric matrix giving, for every pair of
    landmarks, the exact scale at which their edge is witnessed (used by
    \code{\link{build_filtration}} to time every simplex exactly, rather
    than only checking membership at the single scale \code{epsilon}).}
}
}
\description{
Construct a Witness Complex
}
\examples{
points <- matrix(rnorm(200), ncol = 2)
witness_complex <- WitnessComplex(points, landmarks = 15, epsilon = 0.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Distance.R
\name{diagram_matching}
\alias{diagram_matching}
\title{Optimal point matching between two persistence diagrams}
\usage{
diagram_matching(
  df1,
  df2,
  dimension,
  distance = c("wasserstein", "bottleneck"),
  p = 2,
  ground = NULL
)
}
\arguments{
\item{df1, df2}{Persistence diagram data frames with columns \code{dim},
\code{birth}, \code{death}.}

\item{dimension}{Homological dimension to compare.}

\item{distance}{Which distance's optimal matching to compute,
\code{"wasserstein"} (default) or \code{"bottleneck"}.}

\item{p}{Wasserstein order, used only when \code{distance = "wasserstein"}.}

\item{ground}{Ground metric on the birth-death plane, \code{"L2"} or
\code{"Linf"}. Defaults to each distance's own default (\code{"L2"} for
Wasserstein, \code{"Linf"} for bottleneck) when left \code{NULL}.}
}
\value{
A list with \code{distance} (the matching's cost, equal to what
  \code{\link{wasserstein_distance}}/\code{\link{bottleneck_distance}}
  would return), \code{matches} (a data frame with columns
  \code{x_birth, x_death, x_essential, y_birth, y_death, y_essential,
  type}; a \code{NA} pair on one side means that row's point matched the
  diagonal), and \code{X}, \code{Y}, \code{essential_X}, \code{essential_Y}
  (the two diagrams' points after the same essential-pair capping
  \code{\link{wasserstein_distance}} uses - see its Details - and which of
  them were essential before capping).
}
\description{
Computes the same optimal matching that
\code{\link{wasserstein_distance}}/\code{\link{bottleneck_distance}}
reduce to internally, and returns it decoded into a readable table
instead of just the distance - what \code{\link{plot_matching}} draws.
Every point of both diagrams appears in exactly one row: matched to a
point of the other diagram (\code{type = "real-real"}), or matched to the
diagonal, i.e. effectively unmatched (\code{type = "x-diagonal"} for a
point of \code{df1}, \code{"y-diagonal"} for a point of \code{df2}).
}
\examples{
df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
diagram_matching(df1, df2, dimension = 0, distance = "wasserstein", p = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComplexUtils.R
\name{min_enclosing_ball}
\alias{min_enclosing_ball}
\title{Minimum enclosing ball of a finite point set (Welzl's algorithm)}
\usage{
min_enclosing_ball(points)
}
\arguments{
\item{points}{A numeric matrix, one point per row (at least 1 row).}
}
\value{
A list with \code{center} (numeric vector) and \code{radius}.
}
\description{
Used by \code{\link{CechComplex}}: the Cech complex includes a simplex
\eqn{\sigma} at scale \eqn{\epsilon} exactly when the balls of radius
\eqn{\epsilon} centered at its vertices have a common point, which happens
if and only if the minimum enclosing ball of \eqn{\sigma}'s vertices has
radius at most \eqn{\epsilon} (this is the standard reduction used e.g. by
GUDHI's Cech complex; see Cavanna, Jahanseir and Sheehy (2017)).
}
\details{
Randomized, expected linear-time in the number of points; the point sets
passed in here are the vertices of a single simplex, so they are always
small.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Betti.R
\name{betti_number}
\alias{betti_number}
\title{Safely compute the rank of a sparse matrix}
\usage{
betti_number(simplices, bound_dim, tol = NULL)
}
\arguments{
\item{simplices}{A list of simplices representing the simplicial complex.}

\item{bound_dim}{The dimension of the boundary to compute the Betti number for.}

\item{tol}{Optional numerical tolerance to pass to \code{rankMatrix()}.}
}
\value{
An integer representing the rank of the matrix.
}
\description{
This helper function wraps \code{Matrix::rankMatrix()} to safely handle empty matrices (i.e., with 0 rows or columns).
}
\examples{
simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))
betti_number(simplices, 0, tol=0.1)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComplexUtils.R
\name{circumsphere}
\alias{circumsphere}
\title{Circumsphere of an affinely independent point set}
\usage{
circumsphere(points)
}
\arguments{
\item{points}{A numeric matrix, one point per row (must be affinely
independent, i.e. "general position" - see Otter et al. (2017)).}
}
\value{
A list with \code{center} (numeric vector) and \code{radius}.
}
\description{
The unique sphere through \code{points} whose center lies in their affine
hull - i.e. the minimal-radius sphere with all of \code{points} on its
boundary. Used by \code{\link{AlphaComplex}}/\code{\link{DelaunayComplex}}
to compute alpha values (unlike \code{\link{min_enclosing_ball}}, which
minimizes radius over ALL enclosing balls, this fixes every point to lie
exactly on the sphere - the two coincide only for simplices whose
circumcenter is not "hidden" inside the simplex in a way that admits a
smaller covering ball).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComplexUtils.R
\name{pairwise_dist}
\alias{pairwise_dist}
\title{Pairwise Euclidean distance matrix}
\usage{
pairwise_dist(points, query = NULL)
}
\arguments{
\item{points}{A numeric matrix, one point per row.}

\item{query}{Optional second numeric matrix; if supplied, returns the
\code{nrow(points)} x \code{nrow(query)} cross-distance matrix instead of
the full pairwise matrix of \code{points} with itself.}
}
\value{
A numeric distance matrix.
}
\description{
Pairwise Euclidean distance matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Persistence.R
\name{restrict_filtration}
\alias{restrict_filtration}
\title{Restrict a filtration list to simplices up to a given dimension}
\usage{
restrict_filtration(filist, max_dimension)
}
\arguments{
\item{filist}{A filtration list, as produced by
\code{\link{build_filtration}}/\code{\link{build_flood_filtration}}.}

\item{max_dimension}{Maximum simplex dimension to keep (\code{0} = vertices
only, \code{1} = vertices + edges, etc.).}
}
\value{
A filtration list containing only the entries with dimension
  \code{<= max_dimension}, in the same relative order.
}
\description{
Drops every simplex of dimension greater than \code{max_dimension} from a
filtration list; everything else (order, ties, \code{t} values) is left
untouched.
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
restrict_filtration(filtration, max_dimension = 1) # vertices + edges only
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Filtration.R
\name{build_filtration}
\alias{build_filtration}
\title{Build a filtration from a point cloud, for any of five complex types}
\usage{
build_filtration(
  points,
  method = c("VR", "Delaunay", "Alpha", "Cech", "Witness"),
  eps_max = NULL,
  landmarks = NULL,
  nu = 1,
  max_dimension = NULL
)
}
\arguments{
\item{points}{A numeric matrix or data.frame, one point per row.}

\item{method}{One of \code{"VR"}, \code{"Delaunay"}, \code{"Alpha"},
\code{"Cech"}, \code{"Witness"}.}

\item{eps_max}{Maximum scale. Required for \code{"VR"}, \code{"Cech"} and
\code{"Witness"}; for \code{"Alpha"} it caps the alpha value (defaults
to \code{Inf}, i.e. no cap); ignored for \code{"Delaunay"} (which has no
scale parameter - see \code{\link{DelaunayComplex}}).}

\item{landmarks}{For \code{method = "Witness"} only: either an integer
(number of landmarks to select via \code{\link{generate_landmarks}}) or
an integer vector of landmark row indices into \code{points}. Required
for \code{"Witness"}.}

\item{nu}{For \code{method = "Witness"} only: the landmark-distance
parameter passed to \code{\link{WitnessComplex}}. Defaults to \code{1}.}

\item{max_dimension}{Optional maximum homology dimension you intend to
compute persistence for.}
}
\value{
A filtration list, as described above. When \code{max_dimension}
  is set, the list carries it as a \code{"max_dimension"} attribute (see
  above); otherwise the attribute is absent, exactly as before this
  parameter existed.
}
\description{
Build a filtration from a point cloud, for any of five complex types
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
vr_filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
alpha_filtration <- build_filtration(points, method = "Alpha")
delaunay_filtration <- build_filtration(points, method = "Delaunay")
cech_filtration <- build_filtration(points, method = "Cech", eps_max = 0.8)
# cap at H1: persistence_pairs()/etc. need no max_dimension of their own
vr_capped <- build_filtration(points, method = "VR", eps_max = 1.2,
                               max_dimension = 1)
pairs <- persistence_pairs(vr_capped)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EulerCharacteristic.R
\name{euler_characteristic}
\alias{euler_characteristic}
\title{Compute the Euler characteristic \eqn{\chi} of a simplicial complex}
\usage{
euler_characteristic(simplices, tol)
}
\arguments{
\item{simplices}{A list of simplices (each a numeric vector).}

\item{tol}{Optional numerical tolerance to pass to \code{rankMatrix()}.}
}
\value{
An integer representing the Euler characteristic \eqn{\chi}.
}
\description{
Compute the Euler characteristic \eqn{\chi} of a simplicial complex
}
\details{
The Euler characteristic is computed as:
\deqn{\chi = \sum_{k=0}^{k_{\max}} (-1)^k \beta_k}
where \eqn{\beta_k} is the \eqn{k}th Betti number, and \eqn{k_{\max}} is the highest dimension of any simplex in the complex.

Interpretation of values:
\itemize{
  \item \eqn{\chi = 2}: Sphere-like surfaces
  \item \eqn{\chi = 1}: Disk-like spaces
  \item \eqn{\chi = 0}: Torus-like or circle-like spaces
  \item \eqn{\chi < 0}: Surfaces with multiple handles or genus
}
}
\examples{
simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))
euler_characteristic(simplices, tol=0.1)
}
\seealso{
\code{\link{betti_number}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Boundary.R
\name{boundary}
\alias{boundary}
\title{Compute the boundary operator for a simplicial complex}
\usage{
boundary(simplices, bound_dim)
}
\arguments{
\item{simplices}{A list of simplices (each a numeric vector).}

\item{bound_dim}{The dimension k of the boundary operator \eqn{\partial_{k}}.}
}
\value{
A sparse matrix representing \eqn{\partial_{k}}.
}
\description{
Compute the boundary operator for a simplicial complex
}
\details{
\deqn{\partial_k \sigma = \sum_i (-1)^i [v_0 v_1 \ldots \hat{v}_i \ldots v_k]}
}
\examples{
simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))
boundary(simplices, 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CubicalComplex.R
\name{build_cubical_filtration}
\alias{build_cubical_filtration}
\title{Build a cubical (grid) filtration from an image}
\usage{
build_cubical_filtration(image, superlevel = FALSE)
}
\arguments{
\item{image}{A numeric matrix (grid of pixel/voxel values, e.g. a
grayscale image with values in \code{[0, 255]}).}

\item{superlevel}{If \code{TRUE}, filters by decreasing value instead
(equivalent to negating \code{image} first) - useful for tracking
bright structures shrinking rather than dark structures growing.}
}
\value{
A filtration list: one \code{list(simplex = <integer vector of
  pixel ids>, t = <numeric>)} per simplex, sorted by (t, dimension,
  lexicographic order) - the same format as
  \code{\link{build_filtration}}.
}
\description{
Build a cubical (grid) filtration from an image
}
\examples{
# a ring (value 1) around a hole (value 5) around a background (value 9):
# the hole is born once the ring closes and dies once its center fills in
image <- matrix(c(
  9, 9, 9, 9, 9,
  9, 1, 1, 1, 9,
  9, 1, 5, 1, 9,
  9, 1, 1, 1, 9,
  9, 9, 9, 9, 9
), nrow = 5, byrow = TRUE)
filtration <- build_cubical_filtration(image)
pairs <- persistence_pairs(filtration)
pairs[pairs$dim == 1, ] # one H1 bar: birth = 1 (ring closes), death = 5
plot_persistence(pairs)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComplexUtils.R
\name{simplices_to_filtration}
\alias{simplices_to_filtration}
\title{Expand maximal simplices into a sorted filtration list}
\usage{
simplices_to_filtration(maximal_simplices, scale_fn, max_dimension = NULL)
}
\arguments{
\item{maximal_simplices}{A list of integer vectors (the maximal simplices).}

\item{scale_fn}{A function taking one simplex (integer vector) and
returning its filtration time.}

\item{max_dimension}{Optional integer cap. If supplied, faces of dimension
greater than \code{max_dimension} are never generated in the first
place - this is a structural cap on \code{kmax}, evaluated BEFORE
\code{\link{faces}} is called, not a post-hoc filter. Callers that need
the persistence "+1 trick" (see \code{\link{restrict_filtration}}) are
responsible for passing \code{max_dimension + 1} here, not
\code{max_dimension} itself - this function does not know about that
convention.}
}
\value{
A filtration list: one \code{list(simplex = <integer vector>, t =
  <numeric>)} per face, sorted by (t, dimension, lexicographic order).
}
\description{
Shared filtration-assembly step used by every \code{build_filtration()}
method: take a set of maximal simplices, generate every face of every
dimension via \code{\link{faces}}, assign each face a filtration time via
\code{scale_fn}, and sort by (time, dimension, lexicographic order) - the
same convention used throughout the package (see \code{\link{faces}} for
what "lexicographic order" means here).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Persistence.R
\name{persistence_pairs}
\alias{persistence_pairs}
\title{Persistence pairs via sparse boundary reduction (for large filtrations)}
\usage{
persistence_pairs(filist, max_dimension = NULL)
}
\arguments{
\item{filist}{Filtration list, each element includes simplex and time.}

\item{max_dimension}{Optional maximum homology dimension to report
(\code{0} = H0 only, \code{1} = H0 and H1, etc.). When set, one extra
dimension is kept internally so dimension-\code{max_dimension} classes
still get their correct death time from their true (max_dimension+1)
killers, and only that extra dimension is dropped from the returned
pairs - see \code{\link{restrict_filtration}}'s Details for why a plain
truncation would be wrong. Leave \code{NULL} (default) to fall back to
\code{filist}'s own \code{"max_dimension"} attribute (set automatically
when \code{filist} came from \code{\link{build_filtration}} with
\code{max_dimension} supplied there), or to report every dimension
present in \code{filist} if that attribute is also absent.}
}
\value{
A data frame with columns: dim, birth, and death.
}
\description{
Same standard column-reduction algorithm as
\code{boundary_info()} + \code{extract_persistence_pairs()}, but each
column is stored as a sorted integer vector over GF(2) instead of a row of
a dense n x n matrix. Memory drops from O(n^2) to O(total number of
non-zeros), which is what makes filtrations with thousands to hundreds of
thousands of simplices (e.g. Flood or large Vietoris-Rips complexes)
feasible. Output is identical to the dense pipeline.
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
pairs <- persistence_pairs(filtration)
pairs_h0_only <- persistence_pairs(filtration, max_dimension = 0)
# or bake the cap into the filtration itself, and drop the argument here:
capped <- build_filtration(points, method = "VR", eps_max = 1.2,
                            max_dimension = 0)
pairs_h0_only2 <- persistence_pairs(capped)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PersistenceLandscape.R
\name{persistence_landscape}
\alias{persistence_landscape}
\title{Compute the persistence landscape of a persistence diagram}
\usage{
persistence_landscape(
  df,
  dimension = 0,
  k_max = NULL,
  resolution = 500,
  t_range = NULL
)
}
\arguments{
\item{df}{A persistence diagram data frame with columns \code{dim},
\code{birth}, \code{death} (e.g. the output of
\code{extract_persistence_pairs()}).}

\item{dimension}{Homological dimension to extract the landscape for.}

\item{k_max}{Number of landscape levels \eqn{\lambda(1,\cdot), \dots,
\lambda(k_{max},\cdot)} to return. Defaults to all levels supported by
the diagram (the number of finite birth-death pairs in \code{dimension});
levels beyond that are identically zero.}

\item{resolution}{Number of grid points used to discretize each landscape
function.}

\item{t_range}{Optional \code{c(t_min, t_max)} grid range. Defaults to
\code{c(min(birth), max(death))} over the selected pairs.}
}
\value{
A data frame with columns \code{t}, \code{k}, \code{value}: the
  discretized landscape functions, one row per (grid point, level) pair.
}
\description{
Converts a persistence diagram data frame (the direct output of
\code{extract_persistence_pairs()}, \code{persistence_pairs()}, or
\code{flood_persistence()}) into its persistence landscape: a collection of
continuous, piecewise-linear functions \eqn{\lambda(k, \cdot)}, obtained by
overlaying the "tent" function of every birth-death pair and, at each time
\code{t}, taking the \eqn{k}th largest tent value (Bubenik, 2015; Chazal
and Michel, 2021, Section 5.4).
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
res <- boundary_info(filtration)
pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
landscape <- persistence_landscape(pairs, dimension = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotPD.R
\name{plot_persistence}
\alias{plot_persistence}
\title{Plot Persistence Diagram}
\usage{
plot_persistence(df)
}
\arguments{
\item{df}{Dataframe from plot_persistence.}
}
\value{
A ggplot2 object representing the persistence diagram.
}
\description{
Plot Persistence Diagram
}
\examples{
points <- matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2)
filtration <- build_filtration(points, method = "VR", eps_max = 1.2)
res <- boundary_info(filtration)
pairs <- extract_persistence_pairs(filtration, res$last_1, res$pivot_owner)
plot_persistence(pairs)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Faces.R
\name{faces}
\alias{faces}
\title{Generate all unique faces of a given dimension from simplices}
\usage{
faces(simplices, target_dim)
}
\arguments{
\item{simplices}{A list of simplices (each a numeric vector).}

\item{target_dim}{The target dimension \eqn{k} for the faces (e.g., \code{0} for vertices, \code{1} for edges, etc.).}
}
\value{
A list of faces (each a numeric vector) of dimension \code{target_dim}.
}
\description{
Generate all unique faces of a given dimension from simplices
}
\details{
The function generates all possible subsets (combinations) of each simplex, removes duplicates,
and filters them to only include those of length \code{target_dim + 1}.

For example, a 2-simplex \code{c(1, 2, 3)} has three 1-dimensional faces (edges):
\code{c(1,2)}, \code{c(1,3)}, and \code{c(2,3)}, and three 0-dimensional faces (vertices):
\code{1}, \code{2}, and \code{3}.
}
\examples{
simplices <- list(c(1, 2), c(3, 4), c(2, 1, 3), c(4, 2))
faces(simplices, target_dim=0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FloodComplex.R
\name{flood_persistence}
\alias{flood_persistence}
\title{Persistence pairs via sparse boundary reduction}
\usage{
flood_persistence(filist, max_dimension = NULL)
}
\arguments{
\item{filist}{A filtration list (from \code{build_flood_filtration} or
\code{build_filtration}).}

\item{max_dimension}{Optional maximum homology dimension to report. When
set, one extra dimension is kept internally so dimension-\code{max_dimension}
classes still get correct death times from their true killers, and only
that extra dimension is dropped from the output - see
\code{\link{restrict_filtration}}'s Details, and
\code{\link{persistence_pairs}} which applies the same correction.
Leave \code{NULL} (default) to report every dimension present.}
}
\value{
A data frame with columns \code{dim}, \code{birth}, \code{death}.
}
\description{
Computes persistence pairs from a filtration list with the standard
column-reduction algorithm, but on sparse columns (integer index vectors
over GF(2)) instead of a dense matrix. Produces the same output format as
\code{extract_persistence_pairs(filist, res$last_1, res$pivot_owner)} while
scaling to the much larger complexes produced by
\code{\link{build_flood_filtration}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FloodComplex.R
\name{as_filtration}
\alias{as_filtration}
\title{Convert a flood_complex object to a filtration list}
\usage{
as_filtration(fc)
}
\arguments{
\item{fc}{A \code{"flood_complex"} object.}
}
\value{
A filtration list compatible with \code{boundary_info}.
}
\description{
Convert a flood_complex object to a filtration list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComplexDistance.R
\name{complex_distance}
\alias{complex_distance}
\title{Wasserstein/bottleneck distance between two point clouds, via one chosen complex}
\usage{
complex_distance(
  points1,
  points2,
  method = c("VR", "Delaunay", "Alpha", "Cech", "Witness"),
  distance = c("wasserstein", "bottleneck"),
  dimension = 0,
  p = 2,
  ground = NULL,
  eps_max = NULL,
  landmarks = NULL,
  nu = 1
)
}
\arguments{
\item{points1, points2}{Numeric matrices or data.frames, one point per row.}

\item{method}{One of \code{"VR"}, \code{"Delaunay"}, \code{"Alpha"},
\code{"Cech"}, \code{"Witness"} - passed to \code{\link{build_filtration}}.}

\item{distance}{One of \code{"wasserstein"} or \code{"bottleneck"}.}

\item{dimension}{Homological dimension to compare.}

\item{p}{Wasserstein order, used only when \code{distance = "wasserstein"}.}

\item{ground}{Ground metric passed to the chosen distance function
(\code{"L2"} or \code{"Linf"}). Defaults to each distance's own default
(\code{"L2"} for Wasserstein, \code{"Linf"} for bottleneck) if left
\code{NULL}.}

\item{eps_max}{Maximum scale, passed to \code{\link{build_filtration}}.
Required for \code{"VR"}, \code{"Cech"} and \code{"Witness"}; optional
for \code{"Alpha"} (defaults to \code{Inf}); ignored for
\code{"Delaunay"}.}

\item{landmarks}{Passed to \code{\link{build_filtration}} for
\code{"Witness"}; if \code{NULL}, defaults independently for each point
cloud to \code{min(30, nrow(points))} landmarks, with a message.}

\item{nu}{Passed to \code{\link{build_filtration}} for \code{"Witness"}.}
}
\value{
A list with:
\describe{
  \item{distance}{The computed distance (a single number).}
  \item{method, distance_type, dimension}{Echoed back for reference.}
  \item{diagram1, diagram2}{The two persistence diagrams (data frames)
    the distance was computed from.}
}
}
\description{
Picks a single \code{\link{build_filtration}} method, builds the
persistence diagram of each point cloud with it, and compares the two
diagrams with either \code{\link{wasserstein_distance}} or
\code{\link{bottleneck_distance}}. Useful for asking "how different are
these two datasets topologically", holding the complex construction fixed
so the comparison is apples-to-apples.
}
\examples{
set.seed(1)
cloud_a <- matrix(rnorm(24), ncol = 2)
cloud_b <- matrix(rnorm(24), ncol = 2) + 0.1
complex_distance(cloud_a, cloud_b, method = "VR", distance = "bottleneck",
                  dimension = 0, eps_max = 0.6)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FloodComplex.R
\name{generate_landmarks}
\alias{generate_landmarks}
\title{Farthest-Point Sampling of landmarks}
\usage{
generate_landmarks(points, n_lms, start_idx = 1)
}
\arguments{
\item{points}{A numeric matrix (N x d) point cloud.}

\item{n_lms}{Number of landmarks to sample (<= N).}

\item{start_idx}{Index of the starting point. Defaults to 1
(flooder defaults to index 0, i.e. the same first point).}
}
\value{
A list with \code{landmarks} (n_lms x d matrix) and
  \code{indices} (row indices into \code{points}).
}
\description{
Selects \code{n_lms} landmarks from a point cloud via (exact) Farthest-Point
Sampling. Equivalent to \code{flooder::generate_landmarks} (which uses an
approximate bucket-FPS; the exact version below gives the same qualitative
coverage).
}
\examples{
\dontrun{
pts <- matrix(rnorm(2000), ncol = 2)
lms <- generate_landmarks(pts, 50)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Distance.R
\name{wasserstein_distance}
\alias{wasserstein_distance}
\title{Wasserstein distance between two persistence diagrams}
\usage{
wasserstein_distance(df1, df2, dimension = 0, p = 2, ground = c("L2", "Linf"))
}
\arguments{
\item{df1, df2}{Persistence diagram data frames with columns \code{dim},
\code{birth}, \code{death} (e.g. the output of
\code{extract_persistence_pairs()} or \code{persistence_pairs()}).}

\item{dimension}{Homological dimension to compare.}

\item{p}{Wasserstein order (\eqn{p \ge 1}). Defaults to \code{2}.}

\item{ground}{Ground metric on the birth-death plane used to measure the
distance between two points (or a point and the diagonal): \code{"L2"}
(Euclidean, default) or \code{"Linf"} (Chebyshev).}
}
\value{
A single non-negative number, the \eqn{p}-Wasserstein distance.
}
\description{
The \eqn{p}-Wasserstein distance between the points of two persistence
diagrams in a given homological dimension, allowing points to be matched
to the diagonal (Cohen-Steiner, Edelsbrunner, Harer and Mileyko (2010),
"Lipschitz Functions Have \eqn{L_p}-Stable Persistence"). Essential
(death = \code{Inf}) classes are kept rather than dropped - see Details.
Computed exactly via the Hungarian algorithm (\code{clue::solve_LSAP}) on
the augmented assignment problem described in
\code{\link{augmented_cost_matrix}}.
}
\examples{
df1 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 2))
df2 <- data.frame(dim = c(0, 0), birth = c(0, 0), death = c(1, 3))
wasserstein_distance(df1, df2, dimension = 0, p = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComplexUtils.R
\name{build_clique_complex}
\alias{build_clique_complex}
\title{Build the clique (flag) complex of a graph on a fixed vertex set}
\usage{
build_clique_complex(n, edges)
}
\arguments{
\item{n}{Number of vertices.}

\item{edges}{An integer matrix with two columns (or a length-\code{2k}
vector, igraph style), one row per edge, 1-based vertex indices.}
}
\value{
A list with \code{network} (an \code{igraph} object, the
  1-skeleton) and \code{simplices} (a list of integer vectors, the
  vertex sets of the maximal cliques, each sorted).
}
\description{
Shared by \code{\link{VietorisRipsComplex}}, \code{\link{CechComplex}} and
\code{\link{WitnessComplex}}: each of them reduces to "connect vertices
that are close enough (in whatever sense that complex uses), then take the
maximal cliques of that graph as the maximal simplices."
}
\keyword{internal}
